
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gtk/gtkadjustment.h>
#include <gtk/gtkwidget.h>
#include <gdk/gdkkeysyms.h>

#include "defines.h"
#include "memory.h"
#include "common.h"
#include "editor.h"
#include "palette.h"
#include "tiled.h"
#include "jpg.h"
#include "png.h"
#include "bmp.h"
#include "pcx.h"
#include "tga.h"
#include "string.h"


/* internal functions */
static void editor_open(GtkWidget *widget, gpointer data);
static void editor_save(GtkWidget *widget, gpointer data);
static void editor_save_as(GtkWidget *widget, gpointer data);
static void editor_file_save_ok(GtkWidget *widget, gpointer data);
static void editor_file_save_cancel(GtkWidget *widget, gpointer data);
static void editor_file_open_ok(GtkWidget *widget, gpointer data);
static void editor_file_open_cancel(GtkWidget *widget, gpointer data);
static void editor_draw_fill(int x, int y);
static void editor_draw_pick(int x, int y);
static void editor_draw_pen(int x, int y);
static void editor_draw_blur_gaussian(unsigned int x, unsigned int y);
static void editor_draw_blur_normal(unsigned int x, unsigned int y);
static int  editor_need_to_draw(int x, int y);
static void editor_draw_cursor(int x, int y, GtkWidget *widget);
static void editor_undo(GtkWidget *widget, gpointer data);
static void editor_clear(GtkWidget *widget, gpointer data);
static void editor_filters_negate(GtkWidget *widget, gpointer data);
static void editor_filters_blur_normal(GtkWidget *widget, gpointer data);
static void editor_filters_blur_gaussian(GtkWidget *widget, gpointer data);
static void editor_filters_build_alpha(GtkWidget *widget, gpointer data);
static void editor_mirror_lr(GtkWidget *widget, gpointer data);
static void editor_mirror_tb(GtkWidget *widget, gpointer data);
static void editor_copy(GtkWidget *widget, gpointer data);
static void editor_paste(GtkWidget *widget, gpointer data);
static void editor_flip_x(GtkWidget *widget, gpointer data);
static void editor_mirror_and_blend_lin_x(GtkWidget *widget, gpointer data);
static void editor_mirror_and_blend_lin_y(GtkWidget *widget, gpointer data);
static void editor_mirror_and_blend_sto_x(GtkWidget *widget, gpointer data);
static void editor_mirror_and_blend_sto_y(GtkWidget *widget, gpointer data);
static void editor_flip_y(GtkWidget *widget, gpointer data);
static void editor_rotate_left(GtkWidget *widget, gpointer data);
static void editor_rotate_right(GtkWidget *widget, gpointer data);
static void editor_zoom_n(GtkWidget *widget, gpointer data);
static void editor_size_n(GtkWidget *widget, gpointer data);
static void editor_view_n(GtkWidget *widget, gpointer data);
static void editor_scale_n(GtkWidget *widget, gpointer data);
static void editor_tools_n(GtkWidget *widget, gpointer data);
static void editor_shift_left(GtkWidget *widget, gpointer data);
static void editor_shift_right(GtkWidget *widget, gpointer data);
static void editor_shift_up(GtkWidget *widget, gpointer data);
static void editor_shift_down(GtkWidget *widget, gpointer data);
static gint editor_key_press(GtkWidget *widget, GdkEventKey *event);
static gint editor_button_press(GtkWidget *widget, GdkEventButton *event);
static gint editor_button_release(GtkWidget *widget, GdkEventButton *event);
static gint editor_motion_notify(GtkWidget *widget, GdkEventButton *event);
static gint editor_leave_notify(GtkWidget *widget, GdkEventButton *event);
static gboolean editor_draw_area_expose(GtkWidget *widget, GdkEventExpose *event, gpointer user_data);


/* editor window data (GTK) */
char editor_window_title[] = "Tile Driller v1.0";

GtkWidget *editor_window = NULL;
GtkWidget *editor_vbox1 = NULL, *editor_vbox2 = NULL, *editor_hbox1 = NULL;
GtkWidget *editor_menubar = NULL;
GtkWidget *editor_draw_area = NULL;
GtkWidget *editor_file_selection_save = NULL, *editor_file_selection_open = NULL;

GtkAccelGroup *editor_accel_group = NULL;
GtkItemFactory *editor_item_factory = NULL;

GtkItemFactoryEntry editor_menu_items[] = {
  { "/_File",           NULL,          NULL, 0, "<Branch>" },
  { "/File/_Open",      "<control>O",  editor_open, 0, NULL },
  { "/File/_Save",      "<control>S",  editor_save, 0, NULL },
  { "/File/Save As",    NULL,          editor_save_as, 0, NULL },
  { "/_Tools",           NULL,         NULL, 0, "<Branch>" },
  { "/Tools/Blur",       NULL,         NULL, 0, "<Branch>" },
  { "/Tools/Blur/Gaussian", "<alt>1",editor_tools_n, TOOL_MODE_BLUR_GAUSSIAN, "<RadioItem>" },
  { "/Tools/Blur/Normal", "<alt>2",  editor_tools_n, TOOL_MODE_BLUR_NORMAL, "/Tools/Blur/Gaussian" },
  { "/Tools/Fill",       "<alt>3",     editor_tools_n, TOOL_MODE_FILL, "/Tools/Blur/Gaussian" },
  { "/Tools/Pen",        "<alt>4",     editor_tools_n, TOOL_MODE_PEN, "/Tools/Blur/Gaussian" },
  { "/Tools/Pick",       "<alt>5",     editor_tools_n, TOOL_MODE_PICK, "/Tools/Blur/Gaussian" },
  { "/_Edit",            NULL,         NULL, 0, "<Branch>" },
  { "/Edit/Undo",        "<control>Z", editor_undo, 0, NULL },
  { "/Edit/sep1",        NULL,         NULL, 0, "<Separator>" },
  { "/Edit/_Copy",       "<control>C", editor_copy, 0, NULL },
  { "/Edit/Paste",       "<control>V", editor_paste, 0, NULL },
  { "/Edit/sep2",        NULL,         NULL, 0, "<Separator>" },
  { "/Edit/Flip",        NULL,         NULL, 0, "<Branch>" },
  { "/Edit/Flip/X",      NULL,         editor_flip_x, 0, NULL },
  { "/Edit/Flip/Y",      NULL,         editor_flip_y, 0, NULL },
  { "/Edit/Mirror",      NULL,         NULL, 0, "<Branch>" },
  { "/Edit/Mirror/Left->Right",   NULL, editor_mirror_lr, 0, NULL },
  { "/Edit/Mirror/Top->Bottom",   NULL, editor_mirror_tb, 0, NULL },
  { "/Edit/Mirror & Blend",           NULL,         NULL, 0, "<Branch>" },
  { "/Edit/Mirror & Blend/Left->Right (Lin)", NULL, editor_mirror_and_blend_lin_x, 0, NULL },
  { "/Edit/Mirror & Blend/Left->Right (Sto)", NULL, editor_mirror_and_blend_sto_x, 0, NULL },
  { "/Edit/Mirror & Blend/Top->Bottom (Lin)", NULL, editor_mirror_and_blend_lin_y, 0, NULL },
  { "/Edit/Mirror & Blend/Top->Bottom (Sto)", NULL, editor_mirror_and_blend_sto_y, 0, NULL },
  { "/Edit/Rotate",      NULL,         NULL, 0, "<Branch>" },
  { "/Edit/Rotate/Left", NULL,         editor_rotate_left, 0, NULL },
  { "/Edit/Rotate/Right",NULL,         editor_rotate_right, 0, NULL },
  { "/Edit/Scale",       NULL,         NULL, 0, "<Branch>" },
  { "/Edit/Scale/50%",   NULL,         editor_scale_n, 2, NULL },
  { "/Edit/Scale/25%",   NULL,         editor_scale_n, 4, NULL },
  { "/Edit/Scale/12.5%", NULL,         editor_scale_n, 8, NULL },
  { "/Edit/Shift",       NULL,         NULL, 0, "<Branch>" },
  { "/Edit/Shift/Left",  NULL,         editor_shift_left, 0, NULL },
  { "/Edit/Shift/Right", NULL,         editor_shift_right, 0, NULL },
  { "/Edit/Shift/Up",    NULL,         editor_shift_up, 0, NULL },
  { "/Edit/Shift/Down",  NULL,         editor_shift_down, 0, NULL },
  { "/Edit/sep8",        NULL,         NULL, 0, "<Separator>" },
  { "/Edit/Clear",       "<control>K", editor_clear, 0, NULL },
  { "/_Filters",         NULL,         NULL, 0, "<Branch>" },
  { "/Filters/Blur",     NULL,         NULL, 0, "<Branch>" },
  { "/Filters/Blur/Gaussian", NULL,  editor_filters_blur_gaussian, 0, NULL },
  { "/Filters/Blur/Normal",   NULL,  editor_filters_blur_normal, 0, NULL },
  { "/Filters/Build Alpha", NULL,      editor_filters_build_alpha, 0, NULL },
  { "/Filters/Negate",   NULL,         editor_filters_negate, 0, NULL },
  { "/_Zoom",            NULL,         NULL, 0, "<Branch>" },
  { "/Zoom/100%",        NULL,         editor_zoom_n, 1, "<RadioItem>" },
  { "/Zoom/200%",        NULL,         editor_zoom_n, 2, "/Zoom/100%" },
  { "/Zoom/400%",        NULL,         editor_zoom_n, 4, "/Zoom/100%" },
  { "/Zoom/600%",        NULL,         editor_zoom_n, 6, "/Zoom/100%" },
  { "/Zoom/1000%",       NULL,         editor_zoom_n, 10, "/Zoom/100%" },
  { "/Zoom/1500%",       NULL,         editor_zoom_n, 15, "/Zoom/100%" },
  { "/Zoom/2000%",       NULL,         editor_zoom_n, 20, "/Zoom/100%" },
  { "/_Size",            NULL,         NULL, 0, "<Branch>" },
  { "/Size/8x8",         NULL,         editor_size_n, 8, "<RadioItem>" },
  { "/Size/16x16",       NULL,         editor_size_n, 16, "/Size/8x8" },
  { "/Size/32x32",       NULL,         editor_size_n, 32, "/Size/8x8" },
  { "/Size/64x64",       NULL,         editor_size_n, 64, "/Size/8x8" },
  { "/Size/128x128",     NULL,         editor_size_n, 128, "/Size/8x8" },
  { "/Size/256x256",     NULL,         editor_size_n, 256, "/Size/8x8" },
  { "/Size/512x512",     NULL,         editor_size_n, 512, "/Size/8x8" },
  { "/Size/1024x1024",   NULL,         editor_size_n, 1024, "/Size/8x8" },
  { "/_View",            NULL,         NULL, 0, "<Branch>" },
  { "/View/RGBA",        "<alt>6",     editor_view_n, VIEW_MODE_RGBA, "<RadioItem>" },
  { "/View/RGB",         "<alt>7",     editor_view_n, VIEW_MODE_RGB, "/View/RGBA" },
  { "/View/Alpha",       "<alt>8",     editor_view_n, VIEW_MODE_ALPHA, "/View/RGBA" },
};

/* editor data */
int editor_zoom = 10;
int editor_dx = 16, editor_dy = 16;
unsigned char *editor_data = NULL, *editor_view = NULL;

int editor_view_mode = VIEW_MODE_RGBA;

int editor_tile = 0;

/* copy buffer */
int editor_copy_dx = 0, editor_copy_dy = 0;
unsigned char *editor_copy_data = NULL;

/* mouse */
int editor_mouse_button = OFF;
int editor_mouse_button_mode=FG;

/* tools and drawing */
unsigned char editor_pen_r = 0, editor_pen_g = 0, editor_pen_b = 0, editor_pen_a = 255;
unsigned char editor_pen_bg_r = 255, editor_pen_bg_g = 255, editor_pen_bg_b = 255, editor_pen_bg_a = 255;

unsigned char editor_fill_r, editor_fill_g, editor_fill_b, editor_fill_a;
int editor_tool_mode = TOOL_MODE_PEN;

/* undo */
unsigned char *editor_undo_data = NULL, *editor_undo_tmp = NULL;
int editor_undo_dx = 0, editor_undo_dy = 0;

/* file data */
char editor_file_name[256], editor_file_path[256];
int editor_file_status = NO, editor_file_name_status = NO;

/* need to save? */
int editor_need_to_save = NO;


int editor_window_init(void) {

  /* window */
  editor_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  editor_window_set_title();

  /* menu */
  editor_accel_group = gtk_accel_group_new();
  editor_item_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", editor_accel_group);
  gtk_item_factory_create_items(editor_item_factory, sizeof(editor_menu_items)/sizeof(editor_menu_items[0]), editor_menu_items, NULL);
  gtk_window_add_accel_group(GTK_WINDOW(editor_window), editor_accel_group);
  editor_menubar = gtk_item_factory_get_widget(editor_item_factory, "<main>");

  /* drawing area */
  editor_draw_area = gtk_drawing_area_new();
  gtk_drawing_area_size(GTK_DRAWING_AREA(editor_draw_area), editor_dx*editor_zoom, editor_dy*editor_zoom);
  gtk_widget_add_events(editor_draw_area,
			GDK_KEY_PRESS_MASK | GDK_LEAVE_NOTIFY_MASK | GDK_BUTTON_PRESS_MASK |
                        GDK_BUTTON_RELEASE_MASK | GDK_KEY_RELEASE_MASK | GDK_POINTER_MOTION_MASK);

  /* boxes */
  editor_vbox1 = gtk_vbox_new(FALSE, 0);
  editor_vbox2 = gtk_vbox_new(FALSE, 0);
  editor_hbox1 = gtk_hbox_new(FALSE, 0);
  gtk_container_border_width(GTK_CONTAINER(editor_vbox1), 1);
  gtk_container_border_width(GTK_CONTAINER(editor_vbox2), 1);
  gtk_container_border_width(GTK_CONTAINER(editor_hbox1), 1);

  /* place boxes and widgets */
  gtk_container_add(GTK_CONTAINER(editor_window), editor_vbox1);
  gtk_box_pack_start(GTK_BOX(editor_vbox1), editor_menubar, FALSE, FALSE, 0);
  gtk_container_add(GTK_CONTAINER(editor_vbox1), editor_hbox1);
  gtk_box_pack_start(GTK_BOX(editor_hbox1), editor_vbox2, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(editor_vbox2), editor_draw_area, TRUE, FALSE, 0);

  /* file selections */
  editor_file_selection_save = gtk_file_selection_new("Save PNG Image");
  editor_file_selection_open = gtk_file_selection_new("Open BMP/JPG/PCX/PNG/TGA Image");

  /* signals */
  gtk_signal_connect(GTK_OBJECT(editor_draw_area), "expose_event", GTK_SIGNAL_FUNC(editor_draw_area_expose), NULL);
  gtk_signal_connect(GTK_OBJECT(editor_draw_area), "button_press_event", GTK_SIGNAL_FUNC(editor_button_press), NULL);
  gtk_signal_connect(GTK_OBJECT(editor_draw_area), "button_release_event", GTK_SIGNAL_FUNC(editor_button_release), NULL);
  gtk_signal_connect(GTK_OBJECT(editor_draw_area), "motion_notify_event", GTK_SIGNAL_FUNC(editor_motion_notify), NULL);
  gtk_signal_connect(GTK_OBJECT(editor_draw_area), "leave_notify_event", GTK_SIGNAL_FUNC(editor_leave_notify), NULL);
  gtk_signal_connect(GTK_OBJECT(editor_window), "key_press_event", GTK_SIGNAL_FUNC(editor_key_press), NULL);
  gtk_signal_connect(GTK_OBJECT(editor_window), "delete_event", GTK_SIGNAL_FUNC(common_gtk_exit), NULL);
  gtk_signal_connect(GTK_OBJECT(editor_window), "destroy", GTK_SIGNAL_FUNC(common_gtk_exit), NULL);

  gtk_signal_connect(GTK_OBJECT(editor_file_selection_save), "delete_event", GTK_SIGNAL_FUNC(common_gtk_hide_widget), NULL);
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(editor_file_selection_save)->ok_button), "clicked", GTK_SIGNAL_FUNC(editor_file_save_ok), NULL);
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(editor_file_selection_save)->cancel_button), "clicked", GTK_SIGNAL_FUNC(editor_file_save_cancel), NULL);

  gtk_signal_connect(GTK_OBJECT(editor_file_selection_open), "delete_event", GTK_SIGNAL_FUNC(common_gtk_hide_widget), NULL);
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(editor_file_selection_open)->ok_button), "clicked", GTK_SIGNAL_FUNC(editor_file_open_ok), NULL);
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(editor_file_selection_open)->cancel_button), "clicked", GTK_SIGNAL_FUNC(editor_file_open_cancel), NULL);

  gtk_window_set_resizable(GTK_WINDOW(editor_window), FALSE);

  editor_file_name[0] = 0;

  return SUCCEEDED;
}


static gint editor_key_press(GtkWidget *widget, GdkEventKey *event) {

  if (event->keyval == GDK_Right) {
    /* move one tile right */
    if (memory_x < memory_dx - editor_dx) {
      memory_window_plain_block_refresh();
      memory_x += editor_dx;
      if (memory_compute_tile_position(memory_x*memory_zoom, memory_y*memory_zoom) == SUCCEEDED) {
	editor_window_set_title();
	memory_copy_data_out(editor_data, memory_x, memory_y, editor_dx, editor_dy);
	editor_window_refresh();
	memory_window_block_refresh();
	tiled_window_refresh();
      }
    }
  }
  else if (event->keyval == GDK_Left) {
    /* move one tile left */
    if (memory_x >= editor_dx) {
      memory_window_plain_block_refresh();
      memory_x -= editor_dx;
      if (memory_compute_tile_position(memory_x*memory_zoom, memory_y*memory_zoom) == SUCCEEDED) {
	editor_window_set_title();
	memory_copy_data_out(editor_data, memory_x, memory_y, editor_dx, editor_dy);
	editor_window_refresh();
	memory_window_block_refresh();
	tiled_window_refresh();
      }
    }
  }
  else if (event->keyval == GDK_Up) {
    /* move one tile up */
    if (memory_y >= editor_dy) {
      memory_window_plain_block_refresh();
      memory_y -= editor_dy;
      if (memory_compute_tile_position(memory_x*memory_zoom, memory_y*memory_zoom) == SUCCEEDED) {
	editor_window_set_title();
	memory_copy_data_out(editor_data, memory_x, memory_y, editor_dx, editor_dy);
	editor_window_refresh();
	memory_window_block_refresh();
	tiled_window_refresh();
      }
    }
  }
  else if (event->keyval == GDK_Down) {
    /* move one tile down */
    if (memory_y < memory_dy - editor_dy) {
      memory_window_plain_block_refresh();
      memory_y += editor_dy;
      if (memory_compute_tile_position(memory_x*memory_zoom, memory_y*memory_zoom) == SUCCEEDED) {
	editor_window_set_title();
	memory_copy_data_out(editor_data, memory_x, memory_y, editor_dx, editor_dy);
	editor_window_refresh();
	memory_window_block_refresh();
	tiled_window_refresh();
      }
    }
  }
  else if (event->keyval == GDK_a) {
    editor_shift_left(NULL, NULL);
  }
  else if (event->keyval == GDK_d) {
    editor_shift_right(NULL, NULL);
  }
  else if (event->keyval == GDK_w) {
    editor_shift_up(NULL, NULL);
  }
  else if (event->keyval == GDK_s) {
    editor_shift_down(NULL, NULL);
  }
  else if (event->keyval == GDK_x) {
    editor_flip_x(NULL, NULL);
  }
  else if (event->keyval == GDK_y) {
    editor_flip_y(NULL, NULL);
  }
  else if (event->keyval == GDK_q) {
    editor_rotate_left(NULL, NULL);
  }
  else if (event->keyval == GDK_e) {
    editor_rotate_right(NULL, NULL);
  }

  return FALSE;
}


int editor_window_refresh(void) {

  if (editor_data == NULL || editor_view == NULL)
    return FAILED;

  common_copy_data_to_view(editor_data, editor_view, editor_dx, editor_dy, editor_zoom, editor_view_mode);

  /* refresh the editor draw area */
  gtk_widget_queue_draw(editor_draw_area);

  return SUCCEEDED;
}


static void editor_copy_to_undo(void) {

  if (editor_undo_dx != editor_dx || editor_undo_dy != editor_dy || editor_undo_data == NULL) {
    editor_undo_data = realloc(editor_undo_data, editor_dx*editor_dy*4);
    editor_undo_tmp = realloc(editor_undo_tmp, editor_dx*editor_dy*4);
    if (editor_undo_data == NULL || editor_undo_tmp == NULL) {
      fprintf(stderr, "EDITOR_COPY_TO_UNDO: Out of memory error.\n");
      return;
    }

    editor_undo_dx = editor_dx;
    editor_undo_dy = editor_dy;
  }

  /* backup the data */
  memcpy(editor_undo_data, editor_data, editor_dx*editor_dy*4);
}


static void editor_undo(GtkWidget *widget, gpointer data) {

  if (editor_undo_dx != editor_dx || editor_undo_dy != editor_dy || editor_undo_data == NULL || editor_undo_tmp == NULL)
    return;

  /* backup the current data */
  memcpy(editor_undo_tmp, editor_data, editor_dx*editor_dy*4);

  /* return the old data */
  memcpy(editor_data, editor_undo_data, editor_dx*editor_dy*4);

  /* store the previous data */
  memcpy(editor_undo_data, editor_undo_tmp, editor_dx*editor_dy*4);

  /* copy to memory */
  memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

  /* refresh the draw areas */
  editor_window_refresh();
  memory_window_block_refresh();
  tiled_window_refresh();
}


static void editor_clear(GtkWidget *widget, gpointer data) {

  int i;


  /* backup the old data */
  editor_copy_to_undo();

  memory_title_reset(YES);

  /* clear the editor window */
  if (editor_view_mode == VIEW_MODE_RGBA) {
    for (i = 0; i < editor_dx*editor_dy*4; i += 4) {
      editor_data[i + 0] = editor_pen_bg_r;
      editor_data[i + 1] = editor_pen_bg_g;
      editor_data[i + 2] = editor_pen_bg_b;
      editor_data[i + 3] = editor_pen_bg_a;
    }
  }
  else if (editor_view_mode == VIEW_MODE_RGB) {
    for (i = 0; i < editor_dx*editor_dy*4; i += 4) {
      editor_data[i + 0] = editor_pen_bg_r;
      editor_data[i + 1] = editor_pen_bg_g;
      editor_data[i + 2] = editor_pen_bg_b;
    }
  }
  else if (editor_view_mode == VIEW_MODE_ALPHA) {
    for (i = 0; i < editor_dx*editor_dy*4; i += 4) {
      editor_data[i + 3] = editor_pen_bg_a;
    }
  }

  /* copy to memory */
  memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

  /* refresh the draw areas */
  editor_window_refresh();
  memory_window_block_refresh();
  tiled_window_refresh();
}


static void editor_filters_blur_gaussian(GtkWidget *widget, gpointer data) {

  unsigned int x, y, a, b, n, i;
  unsigned char *t;
  float cr, cg, cb, ca;
  float filter[3*3] = { 1.0/36.0, 1.0/9.0, 1.0/36.0,
			1.0/9.0, 4.0/9.0, 1.0/9.0,
			1.0/36.0, 1.0/9.0, 1.0/36.0 };


  t = malloc(editor_dx*editor_dy*4);
  if (t == NULL) {
    fprintf(stderr, "EDITOR_FILTERS_BLUR_GAUSSIAN: Out of memory error.\n");
    return;
  }

  /* backup the old data */
  editor_copy_to_undo();

  memory_title_reset(YES);

  /* blur the data */
  if (editor_view_mode == VIEW_MODE_RGBA) {
    for (y = 0; y < editor_dy; y++) {
      for (x = 0; x < editor_dx; x++) {
	cr = 0;
	cg = 0;
	cb = 0;
	ca = 0;
	for (i = 0, a = 0; a < 3; a++) {
	  for (b = 0; b < 3; b++, i++) {
	    n = (((y+a-1)%editor_dy)*editor_dx + ((x+b-1)%editor_dx))<<2;
	    cr += editor_data[n + 0]*filter[i];
	    cg += editor_data[n + 1]*filter[i];
	    cb += editor_data[n + 2]*filter[i];
	    ca += editor_data[n + 3]*filter[i];
	  }
	}
	n = (y*editor_dx + x)<<2;
	t[n + 0] = cr;
	t[n + 1] = cg;
	t[n + 2] = cb;
	t[n + 3] = ca;
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_RGB) {
    for (y = 0; y < editor_dy; y++) {
      for (x = 0; x < editor_dx; x++) {
	cr = 0;
	cg = 0;
	cb = 0;
	for (i = 0, a = 0; a < 3; a++) {
	  for (b = 0; b < 3; b++, i++) {
	    n = (((y+a-1)%editor_dy)*editor_dx + ((x+b-1)%editor_dx))<<2;
	    cr += editor_data[n + 0]*filter[i];
	    cg += editor_data[n + 1]*filter[i];
	    cb += editor_data[n + 2]*filter[i];
	  }
	}
	n = (y*editor_dx + x)<<2;
	t[n + 0] = cr;
	t[n + 1] = cg;
	t[n + 2] = cb;
	t[n + 3] = editor_data[n + 3];
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_ALPHA) {
    for (y = 0; y < editor_dy; y++) {
      for (x = 0; x < editor_dx; x++) {
	ca = 0;
	for (i = 0, a = 0; a < 3; a++) {
	  for (b = 0; b < 3; b++, i++) {
	    n = (((y+a-1)%editor_dy)*editor_dx + ((x+b-1)%editor_dx))<<2;
	    ca += editor_data[n + 3]*filter[i];
	  }
	}
	n = (y*editor_dx + x)<<2;
	t[n + 0] = editor_data[n + 0];
	t[n + 1] = editor_data[n + 1];
	t[n + 2] = editor_data[n + 2];
	t[n + 3] = ca;
      }
    }
  }

  memcpy(editor_data, t, editor_dx*editor_dy*4);
  free(t);

  /* copy to memory */
  memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

  /* refresh the draw areas */
  editor_window_refresh();
  memory_window_block_refresh();
  tiled_window_refresh();
}


static void editor_filters_blur_normal(GtkWidget *widget, gpointer data) {

  unsigned int x, y, a, b, n;
  unsigned char *t;
  float cr, cg, cb, ca;


  t = malloc(editor_dx*editor_dy*4);
  if (t == NULL) {
    fprintf(stderr, "EDITOR_FILTERS_BLUR_NORMAL: Out of memory error.\n");
    return;
  }

  /* backup the old data */
  editor_copy_to_undo();

  memory_title_reset(YES);

  /* blur the data */
  if (editor_view_mode == VIEW_MODE_RGBA) {
    for (y = 0; y < editor_dy; y++) {
      for (x = 0; x < editor_dx; x++) {
	cr = 0;
	cg = 0;
	cb = 0;
	ca = 0;
	for (a = 0; a < 3; a++) {
	  for (b = 0; b < 3; b++) {
	    n = (((y+a-1)%editor_dy)*editor_dx + ((x+b-1)%editor_dx))<<2;
	    cr += editor_data[n + 0];
	    cg += editor_data[n + 1];
	    cb += editor_data[n + 2];
	    ca += editor_data[n + 3];
	  }
	}
	n = (y*editor_dx + x)<<2;
	t[n + 0] = cr/9.0;
	t[n + 1] = cg/9.0;
	t[n + 2] = cb/9.0;
	t[n + 3] = ca/9.0;
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_RGB) {
    for (y = 0; y < editor_dy; y++) {
      for (x = 0; x < editor_dx; x++) {
	cr = 0;
	cg = 0;
	cb = 0;
	for (a = 0; a < 3; a++) {
	  for (b = 0; b < 3; b++) {
	    n = (((y+a-1)%editor_dy)*editor_dx + ((x+b-1)%editor_dx))<<2;
	    cr += editor_data[n + 0];
	    cg += editor_data[n + 1];
	    cb += editor_data[n + 2];
	  }
	}
	n = (y*editor_dx + x)<<2;
	t[n + 0] = cr/9.0;
	t[n + 1] = cg/9.0;
	t[n + 2] = cb/9.0;
	t[n + 3] = editor_data[n + 3];
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_ALPHA) {
    for (y = 0; y < editor_dy; y++) {
      for (x = 0; x < editor_dx; x++) {
	ca = 0;
	for (a = 0; a < 3; a++) {
	  for (b = 0; b < 3; b++) {
	    n = (((y+a-1)%editor_dy)*editor_dx + ((x+b-1)%editor_dx))<<2;
	    ca += editor_data[n + 3];
	  }
	}
	n = (y*editor_dx + x)<<2;
	t[n + 0] = editor_data[n + 0];
	t[n + 1] = editor_data[n + 1];
	t[n + 2] = editor_data[n + 2];
	t[n + 3] = ca/9.0;
      }
    }
  }

  memcpy(editor_data, t, editor_dx*editor_dy*4);
  free(t);

  /* copy to memory */
  memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

  /* refresh the draw areas */
  editor_window_refresh();
  memory_window_block_refresh();
  tiled_window_refresh();
}


static void editor_filters_negate(GtkWidget *widget, gpointer data) {

  unsigned int x, y, n;


  /* backup the old data */
  editor_copy_to_undo();

  memory_title_reset(YES);

  /* negate the data */
  if (editor_view_mode == VIEW_MODE_RGBA) {
    for (y = 0; y < editor_dy; y++) {
      for (x = 0; x < editor_dx; x++) {
	n = (y*editor_dx + x)<<2;
	editor_data[n + 0] = 255 - editor_data[n + 0];
	editor_data[n + 1] = 255 - editor_data[n + 1];
	editor_data[n + 2] = 255 - editor_data[n + 2];
	editor_data[n + 3] = 255 - editor_data[n + 3];
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_RGB) {
    for (y = 0; y < editor_dy; y++) {
      for (x = 0; x < editor_dx; x++) {
	n = (y*editor_dx + x)<<2;
	editor_data[n + 0] = 255 - editor_data[n + 0];
	editor_data[n + 1] = 255 - editor_data[n + 1];
	editor_data[n + 2] = 255 - editor_data[n + 2];
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_ALPHA) {
    for (y = 0; y < editor_dy; y++) {
      for (x = 0; x < editor_dx; x++) {
	n = (y*editor_dx + x)<<2;
	editor_data[n + 3] = 255 - editor_data[n + 3];
      }
    }
  }

  /* copy to memory */
  memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

  /* refresh the draw areas */
  editor_window_refresh();
  memory_window_block_refresh();
  tiled_window_refresh();
}


static void editor_filters_build_alpha(GtkWidget *widget, gpointer data) {

  unsigned int x, y, n;


  /* backup the old data */
  editor_copy_to_undo();

  memory_title_reset(YES);

  /* build the alpha channel from the colors */
  for (y = 0; y < editor_dy; y++) {
    for (x = 0; x < editor_dx; x++) {
      n = (y*editor_dx + x)<<2;
      if (editor_data[n + 0] == editor_pen_r && editor_data[n + 1] == editor_pen_g && editor_data[n + 2] == editor_pen_b)
	editor_data[n + 3] = 0;
      else
	editor_data[n + 3] = 255;
    }
  }

  /* copy to memory */
  memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

  /* refresh the draw areas */
  editor_window_refresh();
  memory_window_block_refresh();
  tiled_window_refresh();
}


static void editor_mirror_lr(GtkWidget *widget, gpointer data) {

  unsigned int x, y;


  /* backup the old data */
  editor_copy_to_undo();

  memory_title_reset(YES);

  if (editor_view_mode == VIEW_MODE_RGBA) {
    for (y = 0; y < editor_dy; y++) {
      for (x = 0; x < editor_dx/2; x++) {
	editor_data[(y*editor_dx + (editor_dx-x-1))*4 + 0] = editor_data[(y*editor_dx + x)*4 + 0];
	editor_data[(y*editor_dx + (editor_dx-x-1))*4 + 1] = editor_data[(y*editor_dx + x)*4 + 1];
	editor_data[(y*editor_dx + (editor_dx-x-1))*4 + 2] = editor_data[(y*editor_dx + x)*4 + 2];
	editor_data[(y*editor_dx + (editor_dx-x-1))*4 + 3] = editor_data[(y*editor_dx + x)*4 + 3];
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_RGB) {
    for (y = 0; y < editor_dy; y++) {
      for (x = 0; x < editor_dx/2; x++) {
	editor_data[(y*editor_dx + (editor_dx-x-1))*4 + 0] = editor_data[(y*editor_dx + x)*4 + 0];
	editor_data[(y*editor_dx + (editor_dx-x-1))*4 + 1] = editor_data[(y*editor_dx + x)*4 + 1];
	editor_data[(y*editor_dx + (editor_dx-x-1))*4 + 2] = editor_data[(y*editor_dx + x)*4 + 2];
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_ALPHA) {
    for (y = 0; y < editor_dy; y++) {
      for (x = 0; x < editor_dx/2; x++) {
	editor_data[(y*editor_dx + (editor_dx-x-1))*4 + 3] = editor_data[(y*editor_dx + x)*4 + 3];
      }
    }
  }

  /* copy to memory */
  memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

  /* refresh the draw areas */
  editor_window_refresh();
  memory_window_block_refresh();
  tiled_window_refresh();
}


static void editor_mirror_tb(GtkWidget *widget, gpointer data) {

  unsigned int x, y;


  /* backup the old data */
  editor_copy_to_undo();

  memory_title_reset(YES);

  if (editor_view_mode == VIEW_MODE_RGBA) {
    for (y = 0; y < editor_dy/2; y++) {
      for (x = 0; x < editor_dx; x++) {
	editor_data[((editor_dy-y-1)*editor_dx + x)*4 + 0] = editor_data[(y*editor_dx + x)*4 + 0];
	editor_data[((editor_dy-y-1)*editor_dx + x)*4 + 1] = editor_data[(y*editor_dx + x)*4 + 1];
	editor_data[((editor_dy-y-1)*editor_dx + x)*4 + 2] = editor_data[(y*editor_dx + x)*4 + 2];
	editor_data[((editor_dy-y-1)*editor_dx + x)*4 + 3] = editor_data[(y*editor_dx + x)*4 + 3];
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_RGB) {
    for (y = 0; y < editor_dy/2; y++) {
      for (x = 0; x < editor_dx; x++) {
	editor_data[((editor_dy-y-1)*editor_dx + x)*4 + 0] = editor_data[(y*editor_dx + x)*4 + 0];
	editor_data[((editor_dy-y-1)*editor_dx + x)*4 + 1] = editor_data[(y*editor_dx + x)*4 + 1];
	editor_data[((editor_dy-y-1)*editor_dx + x)*4 + 2] = editor_data[(y*editor_dx + x)*4 + 2];
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_ALPHA) {
    for (y = 0; y < editor_dy/2; y++) {
      for (x = 0; x < editor_dx; x++) {
	editor_data[((editor_dy-y-1)*editor_dx + x)*4 + 3] = editor_data[(y*editor_dx + x)*4 + 3];
      }
    }
  }

  /* copy to memory */
  memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

  /* refresh the draw areas */
  editor_window_refresh();
  memory_window_block_refresh();
  tiled_window_refresh();
}


static void editor_copy(GtkWidget *widget, gpointer data) {

  if (editor_copy_dx != editor_dx || editor_copy_dy != editor_dy) {
    editor_copy_data = realloc(editor_copy_data, editor_dx*editor_dy*4);
    if (editor_copy_data == NULL) {
      fprintf(stderr, "EDITOR_COPY: Out of memory error.\n");
      return;
    }

    editor_copy_dx = editor_dx;
    editor_copy_dy = editor_dy;
  }

  /* copy the data to the copy buffer */
  memcpy(editor_copy_data, editor_data, editor_dx*editor_dy*4);
}


static void editor_paste(GtkWidget *widget, gpointer data) {

  int i;


  if (editor_copy_dx != editor_dx || editor_copy_dy != editor_dy)
    return;

  /* backup the old data */
  editor_copy_to_undo();

  memory_title_reset(YES);

  /* copy the data from the copy buffer */
  if (editor_view_mode == VIEW_MODE_RGBA) {
    for (i = 0; i < editor_dx*editor_dy*4; i += 4) {
      editor_data[i + 0] = editor_copy_data[i + 0];
      editor_data[i + 1] = editor_copy_data[i + 1];
      editor_data[i + 2] = editor_copy_data[i + 2];
      editor_data[i + 3] = editor_copy_data[i + 3];
    }
  }
  else if (editor_view_mode == VIEW_MODE_RGB) {
    for (i = 0; i < editor_dx*editor_dy*4; i += 4) {
      editor_data[i + 0] = editor_copy_data[i + 0];
      editor_data[i + 1] = editor_copy_data[i + 1];
      editor_data[i + 2] = editor_copy_data[i + 2];
    }
  }
  else if (editor_view_mode == VIEW_MODE_ALPHA) {
    for (i = 0; i < editor_dx*editor_dy*4; i += 4) {
      editor_data[i + 0] = editor_copy_data[i + 3];
    }
  }

  memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

  /* refresh the draw areas */
  editor_window_refresh();
  memory_window_block_refresh();
  tiled_window_refresh();
}


static void editor_rotate_right(GtkWidget *widget, gpointer data) {

  unsigned int x, y;


  /* backup the old data */
  editor_copy_to_undo();

  memory_title_reset(YES);

  if (editor_view_mode == VIEW_MODE_RGBA) {
    for (y = 0; y < editor_dx; y++) {
      for (x = 0; x < editor_dy; x++) {
	editor_data[y*editor_dx*4 + x*4 + 0] = editor_undo_data[(editor_dx-x-1)*editor_dx*4 + y*4 + 0];
	editor_data[y*editor_dx*4 + x*4 + 1] = editor_undo_data[(editor_dx-x-1)*editor_dx*4 + y*4 + 1];
	editor_data[y*editor_dx*4 + x*4 + 2] = editor_undo_data[(editor_dx-x-1)*editor_dx*4 + y*4 + 2];
	editor_data[y*editor_dx*4 + x*4 + 3] = editor_undo_data[(editor_dx-x-1)*editor_dx*4 + y*4 + 3];
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_RGB) {
    for (y = 0; y < editor_dx; y++) {
      for (x = 0; x < editor_dy; x++) {
	editor_data[y*editor_dx*4 + x*4 + 0] = editor_undo_data[(editor_dx-x-1)*editor_dx*4 + y*4 + 0];
	editor_data[y*editor_dx*4 + x*4 + 1] = editor_undo_data[(editor_dx-x-1)*editor_dx*4 + y*4 + 1];
	editor_data[y*editor_dx*4 + x*4 + 2] = editor_undo_data[(editor_dx-x-1)*editor_dx*4 + y*4 + 2];
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_ALPHA) {
    for (y = 0; y < editor_dx; y++) {
      for (x = 0; x < editor_dy; x++) {
	editor_data[y*editor_dx*4 + x*4 + 3] = editor_undo_data[(editor_dx-x-1)*editor_dx*4 + y*4 + 3];
      }
    }
  }

  /* copy to memory */
  memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

  /* refresh the draw areas */
  editor_window_refresh();
  memory_window_block_refresh();
  tiled_window_refresh();
}


static void editor_rotate_left(GtkWidget *widget, gpointer data) {

  unsigned int x, y;


  /* backup the old data */
  editor_copy_to_undo();

  memory_title_reset(YES);

  if (editor_view_mode == VIEW_MODE_RGBA) {
    for (y = 0; y < editor_dx; y++) {
      for (x = 0; x < editor_dy; x++) {
	editor_data[y*editor_dx*4 + x*4 + 0] = editor_undo_data[x*editor_dx*4 + (editor_dy-y-1)*4 + 0];
	editor_data[y*editor_dx*4 + x*4 + 1] = editor_undo_data[x*editor_dx*4 + (editor_dy-y-1)*4 + 1];
	editor_data[y*editor_dx*4 + x*4 + 2] = editor_undo_data[x*editor_dx*4 + (editor_dy-y-1)*4 + 2];
	editor_data[y*editor_dx*4 + x*4 + 3] = editor_undo_data[x*editor_dx*4 + (editor_dy-y-1)*4 + 3];
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_RGB) {
    for (y = 0; y < editor_dx; y++) {
      for (x = 0; x < editor_dy; x++) {
	editor_data[y*editor_dx*4 + x*4 + 0] = editor_undo_data[x*editor_dx*4 + (editor_dy-y-1)*4 + 0];
	editor_data[y*editor_dx*4 + x*4 + 1] = editor_undo_data[x*editor_dx*4 + (editor_dy-y-1)*4 + 1];
	editor_data[y*editor_dx*4 + x*4 + 2] = editor_undo_data[x*editor_dx*4 + (editor_dy-y-1)*4 + 2];
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_ALPHA) {
    for (y = 0; y < editor_dx; y++) {
      for (x = 0; x < editor_dy; x++) {
	editor_data[y*editor_dx*4 + x*4 + 3] = editor_undo_data[x*editor_dx*4 + (editor_dy-y-1)*4 + 3];
      }
    }
  }

  /* copy to memory */
  memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

  /* refresh the draw areas */
  editor_window_refresh();
  memory_window_block_refresh();
  tiled_window_refresh();
}


static void editor_mirror_and_blend_lin_x(GtkWidget *widget, gpointer data) {

  unsigned int x, y, a, b;
  float f, c, d, w, v;


  f = 1.0f/(editor_dx/4.0f);

  /* backup the old data */
  editor_copy_to_undo();

  memory_title_reset(YES);

  if (editor_view_mode == VIEW_MODE_RGBA) {
    for (y = 0; y < editor_dy; y++) {
      c = 0;
      for (x = 0; x < editor_dx/2-1; x++) {
	d = 1.0f - c;
	a = (y*editor_dx + (editor_dx-x-1))*4;
	b = (y*editor_dx + x+1)*4;
	w = editor_data[a + 0];
	v = editor_data[b + 0];
	editor_data[a + 0] = w*c + v*d;
	w = editor_data[a + 1];
	v = editor_data[b + 1];
	editor_data[a + 1] = w*c + v*d;
	w = editor_data[a + 2];
	v = editor_data[b + 2];
	editor_data[a + 2] = w*c + v*d;
	w = editor_data[a + 3];
	v = editor_data[b + 3];
	editor_data[a + 3] = w*c + v*d;
	c += f;
	if (c > 1)
	  c = 1;
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_RGB) {
    for (y = 0; y < editor_dy; y++) {
      c = 0;
      for (x = 0; x < editor_dx/2-1; x++) {
	d = 1.0f - c;
	a = (y*editor_dx + (editor_dx-x-1))*4;
	b = (y*editor_dx + x+1)*4;
	w = editor_data[a + 0];
	v = editor_data[b + 0];
	editor_data[a + 0] = w*c + v*d;
	w = editor_data[a + 1];
	v = editor_data[b + 1];
	editor_data[a + 1] = w*c + v*d;
	w = editor_data[a + 2];
	v = editor_data[b + 2];
	editor_data[a + 2] = w*c + v*d;
	c += f;
	if (c > 1)
	  c = 1;
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_ALPHA) {
    for (y = 0; y < editor_dy; y++) {
      c = 0;
      for (x = 0; x < editor_dx/2-1; x++) {
	d = 1.0f - c;
	a = (y*editor_dx + (editor_dx-x-1))*4;
	b = (y*editor_dx + x+1)*4;
	w = editor_data[a + 2];
	v = editor_data[b + 2];
	editor_data[a + 2] = w*c + v*d;
	c += f;
	if (c > 1)
	  c = 1;
      }
    }
  }

  /* copy to memory */
  memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

  /* refresh the draw areas */
  editor_window_refresh();
  memory_window_block_refresh();
  tiled_window_refresh();
}


static void editor_mirror_and_blend_lin_y(GtkWidget *widget, gpointer data) {

  unsigned int x, y, a, b;
  float f, c, d, w, v;


  f = 1.0f/(editor_dy/4.0f);

  /* backup the old data */
  editor_copy_to_undo();

  memory_title_reset(YES);

  if (editor_view_mode == VIEW_MODE_RGBA) {
    for (x = 0; x < editor_dx; x++) {
      c = 0;
      for (y = 0; y < editor_dy/2-1; y++) {
	d = 1.0f - c;
	a = ((editor_dy-y-1)*editor_dx + x)*4;
	b = ((y+1)*editor_dx + x)*4;
	w = editor_data[a + 0];
	v = editor_data[b + 0];
	editor_data[a + 0] = w*c + v*d;
	w = editor_data[a + 1];
	v = editor_data[b + 1];
	editor_data[a + 1] = w*c + v*d;
	w = editor_data[a + 2];
	v = editor_data[b + 2];
	editor_data[a + 2] = w*c + v*d;
	w = editor_data[a + 3];
	v = editor_data[b + 3];
	editor_data[a + 3] = w*c + v*d;
	c += f;
	if (c > 1)
	  c = 1;
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_RGB) {
    for (x = 0; x < editor_dx; x++) {
      c = 0;
      for (y = 0; y < editor_dy/2-1; y++) {
	d = 1.0f - c;
	a = ((editor_dy-y-1)*editor_dx + x)*4;
	b = ((y+1)*editor_dx + x)*4;
	w = editor_data[a + 0];
	v = editor_data[b + 0];
	editor_data[a + 0] = w*c + v*d;
	w = editor_data[a + 1];
	v = editor_data[b + 1];
	editor_data[a + 1] = w*c + v*d;
	w = editor_data[a + 2];
	v = editor_data[b + 2];
	editor_data[a + 2] = w*c + v*d;
	c += f;
	if (c > 1)
	  c = 1;
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_ALPHA) {
    for (x = 0; x < editor_dx; x++) {
      c = 0;
      for (y = 0; y < editor_dy/2-1; y++) {
	d = 1.0f - c;
	a = ((editor_dy-y-1)*editor_dx + x)*4;
	b = ((y+1)*editor_dx + x)*4;
	w = editor_data[a + 3];
	v = editor_data[b + 3];
	editor_data[a + 3] = w*c + v*d;
	c += f;
	if (c > 1)
	  c = 1;
      }
    }
  }

  /* copy to memory */
  memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

  /* refresh the draw areas */
  editor_window_refresh();
  memory_window_block_refresh();
  tiled_window_refresh();
}


static void editor_mirror_and_blend_sto_x(GtkWidget *widget, gpointer data) {

  unsigned int x, y, a, b;
  float c, d, w, v;


  /* backup the old data */
  editor_copy_to_undo();

  memory_title_reset(YES);

  if (editor_view_mode == VIEW_MODE_RGBA) {
    for (y = 0; y < editor_dy; y++) {
      c = 1.0f - (rand()/(double)RAND_MAX/5.0);
      for (x = 0; x < editor_dx/2-1; x++) {
	d = 1.0f - c;
	a = (y*editor_dx + (editor_dx-x-1))*4;
	b = (y*editor_dx + x+1)*4;
	w = editor_data[a + 0];
	v = editor_data[b + 0];
	editor_data[a + 0] = w*d + v*c;
	w = editor_data[a + 1];
	v = editor_data[b + 1];
	editor_data[a + 1] = w*d + v*c;
	w = editor_data[a + 2];
	v = editor_data[b + 2];
	editor_data[a + 2] = w*d + v*c;
	w = editor_data[a + 3];
	v = editor_data[b + 3];
	editor_data[a + 3] = w*d + v*c;
	c = c - (rand()/(double)RAND_MAX/20.0 + 0.1);
	if (c < 0)
	  c = 0;
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_RGB) {
    for (y = 0; y < editor_dy; y++) {
      c = 1.0f - (rand()/(double)RAND_MAX/5.0);
      for (x = 0; x < editor_dx/2-1; x++) {
	d = 1.0f - c;
	a = (y*editor_dx + (editor_dx-x-1))*4;
	b = (y*editor_dx + x+1)*4;
	w = editor_data[a + 0];
	v = editor_data[b + 0];
	editor_data[a + 0] = w*d + v*c;
	w = editor_data[a + 1];
	v = editor_data[b + 1];
	editor_data[a + 1] = w*d + v*c;
	w = editor_data[a + 2];
	v = editor_data[b + 2];
	editor_data[a + 2] = w*d + v*c;
	c = c - (rand()/(double)RAND_MAX/20.0 + 0.1);
	if (c < 0)
	  c = 0;
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_ALPHA) {
    for (y = 0; y < editor_dy; y++) {
      c = 1.0f - (rand()/(double)RAND_MAX/5.0);
      for (x = 0; x < editor_dx/2-1; x++) {
	d = 1.0f - c;
	a = (y*editor_dx + (editor_dx-x-1))*4;
	b = (y*editor_dx + x+1)*4;
	w = editor_data[a + 3];
	v = editor_data[b + 3];
	editor_data[a + 3] = w*d + v*c;
	c = c - (rand()/(double)RAND_MAX/20.0 + 0.1);
	if (c < 0)
	  c = 0;
      }
    }
  }

  /* copy to memory */
  memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

  /* refresh the draw areas */
  editor_window_refresh();
  memory_window_block_refresh();
  tiled_window_refresh();
}


static void editor_mirror_and_blend_sto_y(GtkWidget *widget, gpointer data) {

  unsigned int x, y, a, b;
  float c, d, w, v;


  /* backup the old data */
  editor_copy_to_undo();

  memory_title_reset(YES);

  if (editor_view_mode == VIEW_MODE_RGBA) {
    for (x = 0; x < editor_dx; x++) {
      c = 1.0f - (rand()/(double)RAND_MAX/5.0);
      for (y = 0; y < editor_dy/2-1; y++) {
	d = 1.0f - c;
	a = ((editor_dy-y-1)*editor_dx + x)*4;
	b = ((y+1)*editor_dx + x)*4;
	w = editor_data[a + 0];
	v = editor_data[b + 0];
	editor_data[a + 0] = w*d + v*c;
	w = editor_data[a + 1];
	v = editor_data[b + 1];
	editor_data[a + 1] = w*d + v*c;
	w = editor_data[a + 2];
	v = editor_data[b + 2];
	editor_data[a + 2] = w*d + v*c;
	w = editor_data[a + 3];
	v = editor_data[b + 3];
	editor_data[a + 3] = w*d + v*c;
	c = c - (rand()/(double)RAND_MAX/20.0 + 0.1);
	if (c < 0)
	  c = 0;
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_RGB) {
    for (x = 0; x < editor_dx; x++) {
      c = 1.0f - (rand()/(double)RAND_MAX/5.0);
      for (y = 0; y < editor_dy/2-1; y++) {
	d = 1.0f - c;
	a = ((editor_dy-y-1)*editor_dx + x)*4;
	b = ((y+1)*editor_dx + x)*4;
	w = editor_data[a + 0];
	v = editor_data[b + 0];
	editor_data[a + 0] = w*d + v*c;
	w = editor_data[a + 1];
	v = editor_data[b + 1];
	editor_data[a + 1] = w*d + v*c;
	w = editor_data[a + 2];
	v = editor_data[b + 2];
	editor_data[a + 2] = w*d + v*c;
	c = c - (rand()/(double)RAND_MAX/20.0 + 0.1);
	if (c < 0)
	  c = 0;
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_ALPHA) {
    for (x = 0; x < editor_dx; x++) {
      c = 1.0f - (rand()/(double)RAND_MAX/5.0);
      for (y = 0; y < editor_dy/2-1; y++) {
	d = 1.0f - c;
	a = ((editor_dy-y-1)*editor_dx + x)*4;
	b = ((y+1)*editor_dx + x)*4;
	w = editor_data[a + 3];
	v = editor_data[b + 3];
	editor_data[a + 3] = w*d + v*c;
	c = c - (rand()/(double)RAND_MAX/20.0 + 0.1);
	if (c < 0)
	  c = 0;
      }
    }
  }

  /* copy to memory */
  memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

  /* refresh the draw areas */
  editor_window_refresh();
  memory_window_block_refresh();
  tiled_window_refresh();
}


static void editor_flip_x(GtkWidget *widget, gpointer data) {

  unsigned char c;
  int x, y, a, b, n;


  /* backup the old data */
  editor_copy_to_undo();

  memory_title_reset(YES);

  if (editor_view_mode == VIEW_MODE_RGBA) {
    for (y = 0; y < editor_dy; y++) {
      a = (y*editor_dx)<<2;
      for (x = 0; x < editor_dx/2; x++) {
	b = a + (x<<2);
	n = a + ((editor_dx-1-x)<<2);
	c = editor_data[b + 0];
	editor_data[b + 0] = editor_data[n + 0];
	editor_data[n + 0] = c;
	c = editor_data[b + 1];
	editor_data[b + 1] = editor_data[n + 1];
	editor_data[n + 1] = c;
	c = editor_data[b + 2];
	editor_data[b + 2] = editor_data[n + 2];
	editor_data[n + 2] = c;
	c = editor_data[b + 3];
	editor_data[b + 3] = editor_data[n + 3];
	editor_data[n + 3] = c;
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_RGB) {
    for (y = 0; y < editor_dy; y++) {
      a = (y*editor_dx)<<2;
      for (x = 0; x < editor_dx/2; x++) {
	b = a + (x<<2);
	n = a + ((editor_dx-1-x)<<2);
	c = editor_data[b + 0];
	editor_data[b + 0] = editor_data[n + 0];
	editor_data[n + 0] = c;
	c = editor_data[b + 1];
	editor_data[b + 1] = editor_data[n + 1];
	editor_data[n + 1] = c;
	c = editor_data[b + 2];
	editor_data[b + 2] = editor_data[n + 2];
	editor_data[n + 2] = c;
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_ALPHA) {
    for (y = 0; y < editor_dy; y++) {
      a = (y*editor_dx)<<2;
      for (x = 0; x < editor_dx/2; x++) {
	b = a + (x<<2);
	n = a + ((editor_dx-1-x)<<2);
	c = editor_data[b + 3];
	editor_data[b + 3] = editor_data[n + 3];
	editor_data[n + 3] = c;
      }
    }
  }

  /* copy to memory */
  memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

  /* refresh the draw areas */
  editor_window_refresh();
  memory_window_block_refresh();
  tiled_window_refresh();
}


static void editor_flip_y(GtkWidget *widget, gpointer data) {

  unsigned char c;
  int x, y, a, b, n;


  /* backup the old data */
  editor_copy_to_undo();

  memory_title_reset(YES);

  if (editor_view_mode == VIEW_MODE_RGBA) {
    for (y = 0; y < editor_dy/2; y++) {
      a = (y*editor_dx)<<2;
      b = ((editor_dy-1-y)*editor_dx)<<2;
      for (x = 0; x < editor_dx; x++) {
	n = x<<2;
	c = editor_data[a + n + 0];
	editor_data[a + n + 0] = editor_data[b + n + 0];
	editor_data[b + n + 0] = c;
	c = editor_data[a + n + 1];
	editor_data[a + n + 1] = editor_data[b + n + 1];
	editor_data[b + n + 1] = c;
	c = editor_data[a + n + 2];
	editor_data[a + n + 2] = editor_data[b + n + 2];
	editor_data[b + n + 2] = c;
	c = editor_data[a + n + 3];
	editor_data[a + n + 3] = editor_data[b + n + 3];
	editor_data[b + n + 3] = c;
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_RGB) {
    for (y = 0; y < editor_dy/2; y++) {
      a = (y*editor_dx)<<2;
      b = ((editor_dy-1-y)*editor_dx)<<2;
      for (x = 0; x < editor_dx; x++) {
	n = x<<2;
	c = editor_data[a + n + 0];
	editor_data[a + n + 0] = editor_data[b + n + 0];
	editor_data[b + n + 0] = c;
	c = editor_data[a + n + 1];
	editor_data[a + n + 1] = editor_data[b + n + 1];
	editor_data[b + n + 1] = c;
	c = editor_data[a + n + 2];
	editor_data[a + n + 2] = editor_data[b + n + 2];
	editor_data[b + n + 2] = c;
      }
    }
  }
  else if (editor_view_mode == VIEW_MODE_ALPHA) {
    for (y = 0; y < editor_dy/2; y++) {
      a = (y*editor_dx)<<2;
      b = ((editor_dy-1-y)*editor_dx)<<2;
      for (x = 0; x < editor_dx; x++) {
	n = (x<<2) + 3;
	c = editor_data[a + n];
	editor_data[a + n] = editor_data[b + n];
	editor_data[b + n] = c;
      }
    }
  }

  /* copy to memory */
  memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

  /* refresh the draw areas */
  editor_window_refresh();
  memory_window_block_refresh();
  tiled_window_refresh();
}


static void editor_shift_left(GtkWidget *widget, gpointer data) {

  unsigned char r, g, b, a;
  int x, y, n;


  /* backup the old data */
  editor_copy_to_undo();

  memory_title_reset(YES);

  if (editor_view_mode == VIEW_MODE_RGBA) {
    for (y = 0; y < editor_dy; y++) {
      n = (y*editor_dx)<<2;
      r = editor_data[n + 0];
      g = editor_data[n + 1];
      b = editor_data[n + 2];
      a = editor_data[n + 3];
      for (x = 0; x < editor_dx-1; x++) {
	editor_data[n + x*4 + 0] = editor_data[n + (x+1)*4 + 0];
	editor_data[n + x*4 + 1] = editor_data[n + (x+1)*4 + 1];
	editor_data[n + x*4 + 2] = editor_data[n + (x+1)*4 + 2];
	editor_data[n + x*4 + 3] = editor_data[n + (x+1)*4 + 3];
      }
      editor_data[n + x*4 + 0] = r;
      editor_data[n + x*4 + 1] = g;
      editor_data[n + x*4 + 2] = b;
      editor_data[n + x*4 + 3] = a;
    }
  }
  else if (editor_view_mode == VIEW_MODE_RGB) {
    for (y = 0; y < editor_dy; y++) {
      n = (y*editor_dx)<<2;
      r = editor_data[n + 0];
      g = editor_data[n + 1];
      b = editor_data[n + 2];
      for (x = 0; x < editor_dx-1; x++) {
	editor_data[n + x*4 + 0] = editor_data[n + (x+1)*4 + 0];
	editor_data[n + x*4 + 1] = editor_data[n + (x+1)*4 + 1];
	editor_data[n + x*4 + 2] = editor_data[n + (x+1)*4 + 2];
      }
      editor_data[n + x*4 + 0] = r;
      editor_data[n + x*4 + 1] = g;
      editor_data[n + x*4 + 2] = b;
    }
  }
  else if (editor_view_mode == VIEW_MODE_ALPHA) {
    for (y = 0; y < editor_dy; y++) {
      n = ((y*editor_dx)<<2) + 3;
      a = editor_data[n];
      for (x = 0; x < editor_dx-1; x++) {
	editor_data[n + x*4] = editor_data[n + (x+1)*4];
      }
      editor_data[n + x*4] = a;
    }
  }


  /* copy to memory */
  memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

  /* refresh the draw areas */
  editor_window_refresh();
  memory_window_block_refresh();
  tiled_window_refresh();
}


static void editor_shift_right(GtkWidget *widget, gpointer data) {

  unsigned char r, g, b, a;
  int x, y, n;


  /* backup the old data */
  editor_copy_to_undo();

  memory_title_reset(YES);

  if (editor_view_mode == VIEW_MODE_RGBA) {
    for (y = 0; y < editor_dy; y++) {
      n = (y*editor_dx)<<2;
      r = editor_data[n + (editor_dx-1)*4 + 0];
      g = editor_data[n + (editor_dx-1)*4 + 1];
      b = editor_data[n + (editor_dx-1)*4 + 2];
      a = editor_data[n + (editor_dx-1)*4 + 3];
      for (x = editor_dx-1; x > 0; x--) {
	editor_data[n + x*4 + 0] = editor_data[n + (x-1)*4 + 0];
	editor_data[n + x*4 + 1] = editor_data[n + (x-1)*4 + 1];
	editor_data[n + x*4 + 2] = editor_data[n + (x-1)*4 + 2];
	editor_data[n + x*4 + 3] = editor_data[n + (x-1)*4 + 3];
      }
      editor_data[n + 0] = r;
      editor_data[n + 1] = g;
      editor_data[n + 2] = b;
      editor_data[n + 3] = a;
    }
  }
  else if (editor_view_mode == VIEW_MODE_RGB) {
    for (y = 0; y < editor_dy; y++) {
      n = (y*editor_dx)<<2;
      r = editor_data[n + (editor_dx-1)*4 + 0];
      g = editor_data[n + (editor_dx-1)*4 + 1];
      b = editor_data[n + (editor_dx-1)*4 + 2];
      for (x = editor_dx-1; x > 0; x--) {
	editor_data[n + x*4 + 0] = editor_data[n + (x-1)*4 + 0];
	editor_data[n + x*4 + 1] = editor_data[n + (x-1)*4 + 1];
	editor_data[n + x*4 + 2] = editor_data[n + (x-1)*4 + 2];
      }
      editor_data[n + 0] = r;
      editor_data[n + 1] = g;
      editor_data[n + 2] = b;
    }
  }
  else if (editor_view_mode == VIEW_MODE_ALPHA) {
    for (y = 0; y < editor_dy; y++) {
      n = ((y*editor_dx)<<2) + 3;
      a = editor_data[n + (editor_dx-1)*4];
      for (x = editor_dx-1; x > 0; x--) {
	editor_data[n + x*4] = editor_data[n + (x-1)*4];
      }
      editor_data[n] = a;
    }
  }

  /* copy to memory */
  memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

  /* refresh the draw areas */
  editor_window_refresh();
  memory_window_block_refresh();
  tiled_window_refresh();
}


static void editor_shift_up(GtkWidget *widget, gpointer data) {

  unsigned char r, g, b, a;
  int x, y, n;


  /* backup the old data */
  editor_copy_to_undo();

  memory_title_reset(YES);

  if (editor_view_mode == VIEW_MODE_RGBA) {
    for (x = 0; x < editor_dx; x++) {
      n = x<<2;
      r = editor_data[n + 0];
      g = editor_data[n + 1];
      b = editor_data[n + 2];
      a = editor_data[n + 3];
      for (y = 0; y < editor_dy-1; y++) {
	editor_data[y*editor_dx*4 + n + 0] = editor_data[(y+1)*editor_dx*4 + n + 0];
	editor_data[y*editor_dx*4 + n + 1] = editor_data[(y+1)*editor_dx*4 + n + 1];
	editor_data[y*editor_dx*4 + n + 2] = editor_data[(y+1)*editor_dx*4 + n + 2];
	editor_data[y*editor_dx*4 + n + 3] = editor_data[(y+1)*editor_dx*4 + n + 3];
      }
      editor_data[y*editor_dx*4 + n + 0] = r;
      editor_data[y*editor_dx*4 + n + 1] = g;
      editor_data[y*editor_dx*4 + n + 2] = b;
      editor_data[y*editor_dx*4 + n + 3] = a;
    }
  }
  else if (editor_view_mode == VIEW_MODE_RGB) {
    for (x = 0; x < editor_dx; x++) {
      n = x<<2;
      r = editor_data[n + 0];
      g = editor_data[n + 1];
      b = editor_data[n + 2];
      for (y = 0; y < editor_dy-1; y++) {
	editor_data[y*editor_dx*4 + n + 0] = editor_data[(y+1)*editor_dx*4 + n + 0];
	editor_data[y*editor_dx*4 + n + 1] = editor_data[(y+1)*editor_dx*4 + n + 1];
	editor_data[y*editor_dx*4 + n + 2] = editor_data[(y+1)*editor_dx*4 + n + 2];
      }
      editor_data[y*editor_dx*4 + n + 0] = r;
      editor_data[y*editor_dx*4 + n + 1] = g;
      editor_data[y*editor_dx*4 + n + 2] = b;
    }
  }
  else if (editor_view_mode == VIEW_MODE_ALPHA) {
    for (x = 0; x < editor_dx; x++) {
      n = (x<<2) + 3;
      a = editor_data[n];
      for (y = 0; y < editor_dy-1; y++) {
	editor_data[y*editor_dx*4 + n] = editor_data[(y+1)*editor_dx*4 + n];
      }
      editor_data[y*editor_dx*4 + n] = a;
    }
  }

  /* copy to memory */
  memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

  /* refresh the draw areas */
  editor_window_refresh();
  memory_window_block_refresh();
  tiled_window_refresh();
}


static void editor_shift_down(GtkWidget *widget, gpointer data) {

  unsigned char r, g, b, a;
  int x, y, n;


  /* backup the old data */
  editor_copy_to_undo();

  memory_title_reset(YES);

  if (editor_view_mode == VIEW_MODE_RGBA) {
    for (x = 0; x < editor_dx; x++) {
      n = x<<2;
      r = editor_data[(editor_dy-1)*editor_dx*4 + n + 0];
      g = editor_data[(editor_dy-1)*editor_dx*4 + n + 1];
      b = editor_data[(editor_dy-1)*editor_dx*4 + n + 2];
      a = editor_data[(editor_dy-1)*editor_dx*4 + n + 3];
      for (y = editor_dy-1; y > 0; y--) {
	editor_data[y*editor_dx*4 + n + 0] = editor_data[(y-1)*editor_dx*4 + n + 0];
	editor_data[y*editor_dx*4 + n + 1] = editor_data[(y-1)*editor_dx*4 + n + 1];
	editor_data[y*editor_dx*4 + n + 2] = editor_data[(y-1)*editor_dx*4 + n + 2];
	editor_data[y*editor_dx*4 + n + 3] = editor_data[(y-1)*editor_dx*4 + n + 3];
      }
      editor_data[n + 0] = r;
      editor_data[n + 1] = g;
      editor_data[n + 2] = b;
      editor_data[n + 3] = a;
    }
  }
  else if (editor_view_mode == VIEW_MODE_RGB) {
    for (x = 0; x < editor_dx; x++) {
      n = x<<2;
      r = editor_data[(editor_dy-1)*editor_dx*4 + n + 0];
      g = editor_data[(editor_dy-1)*editor_dx*4 + n + 1];
      b = editor_data[(editor_dy-1)*editor_dx*4 + n + 2];
      for (y = editor_dy-1; y > 0; y--) {
	editor_data[y*editor_dx*4 + n + 0] = editor_data[(y-1)*editor_dx*4 + n + 0];
	editor_data[y*editor_dx*4 + n + 1] = editor_data[(y-1)*editor_dx*4 + n + 1];
	editor_data[y*editor_dx*4 + n + 2] = editor_data[(y-1)*editor_dx*4 + n + 2];
      }
      editor_data[n + 0] = r;
      editor_data[n + 1] = g;
      editor_data[n + 2] = b;
    }
  }
  else if (editor_view_mode == VIEW_MODE_ALPHA) {
    for (x = 0; x < editor_dx; x++) {
      n = (x<<2) + 3;
      a = editor_data[(editor_dy-1)*editor_dx*4 + n];
      for (y = editor_dy-1; y > 0; y--) {
	editor_data[y*editor_dx*4 + n] = editor_data[(y-1)*editor_dx*4 + n];
      }
      editor_data[n] = a;
    }
  }

  /* copy to memory */
  memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

  /* refresh the draw areas */
  editor_window_refresh();
  memory_window_block_refresh();
  tiled_window_refresh();
}


static void editor_scale_n(GtkWidget *widget, gpointer data) {

  /* backup the old data */
  editor_copy_to_undo();

  memory_title_reset(YES);

  /* scale! */
  common_scale((int)data, editor_view_mode, editor_dx, editor_dy, editor_data);

  /* copy to memory */
  memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

  /* refresh the draw areas */
  editor_window_refresh();
  memory_window_block_refresh();
  tiled_window_refresh();
}


int editor_window_set_title(void) {

  char tmp[512];


  if (editor_window == NULL)
    return FAILED;

  /* display the tile number only when snap-to-grid is enabled */
  if (memory_snap_to_grid == YES)
    sprintf(tmp, "TILE: %d | %dx%d", editor_tile, editor_dx, editor_dy);
  else
    sprintf(tmp, "TILE: ? | %dx%d", editor_dx, editor_dy);

  if (editor_view_mode == VIEW_MODE_RGBA)
    strcat(tmp, " | RGBA");
  else if (editor_view_mode == VIEW_MODE_RGB)
    strcat(tmp, " | RGB");
  else if (editor_view_mode == VIEW_MODE_ALPHA)
    strcat(tmp, " | Alpha");

  if (editor_tool_mode == TOOL_MODE_BLUR_GAUSSIAN)
    strcat(tmp, " | Blur (Gaussian)");
  else if (editor_tool_mode == TOOL_MODE_BLUR_NORMAL)
    strcat(tmp, " | Blur (Normal)");
  else if (editor_tool_mode == TOOL_MODE_PEN)
    strcat(tmp, " | Pen");
  else if (editor_tool_mode == TOOL_MODE_PICK)
    strcat(tmp, " | Pick");
  else if (editor_tool_mode == TOOL_MODE_FILL)
    strcat(tmp, " | Fill");

  gtk_window_set_title(GTK_WINDOW(editor_window), tmp);

  return SUCCEEDED;
}


void editor_resize_view(int zoom) {

  editor_view = realloc(editor_view, editor_dx*zoom*editor_dy*zoom*3);
  if (editor_view == NULL) {
    fprintf(stderr, "EDITOR_RESIZE_VIEW: Out of memory error.\n");
    return;
  }

  editor_zoom = zoom;
}


void editor_resize_data(int dx, int dy) {

  editor_data = realloc(editor_data, dx*dy*4);
  if (editor_data == NULL) {
    fprintf(stderr, "EDITOR_RESIZE_DATA: Out of memory error.\n");
    return;
  }

  memset(editor_data, 255, dx*dy*4);

  editor_dx = dx;
  editor_dy = dy;
}


static void editor_zoom_n(GtkWidget *widget, gpointer data) {

  int zoom;


  zoom = (int)data;
  if (editor_zoom == zoom)
    return;

  editor_resize_view(zoom);
  common_copy_data_to_view(editor_data, editor_view, editor_dx, editor_dy, editor_zoom, editor_view_mode);

  gtk_drawing_area_size(GTK_DRAWING_AREA(editor_draw_area), editor_dx*editor_zoom, editor_dy*editor_zoom);
}


static void editor_size_n(GtkWidget *widget, gpointer data) {

  int i;


  i = (int)data;
  if (editor_dx == i && editor_dy == i)
    return;

  /* tile size cannot be larger than the memory size */
  if (memory_dx < i || memory_dy < i) {
    if (editor_dx == 8)
      gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Size/8x8"), TRUE);
    else if (editor_dx == 16)
      gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Size/16x16"), TRUE);
    else if (editor_dx == 32)
      gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Size/32x32"), TRUE);
    else if (editor_dx == 64)
      gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Size/64x64"), TRUE);
    else if (editor_dx == 128)
      gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Size/128x128"), TRUE);
    else if (editor_dx == 256)
      gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Size/256x256"), TRUE);
    else if (editor_dx == 512)
      gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Size/512x512"), TRUE);
    else if (editor_dx == 1024)
      gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(editor_item_factory, "/Size/1024x1024"), TRUE);
    return;
  }

  editor_resize_data(i, i);
  editor_resize_view(editor_zoom);

  if (memory_x + editor_dx > memory_dx)
    memory_x = memory_dx - editor_dx;
  if (memory_y + editor_dy > memory_dy)
    memory_y = memory_dy - editor_dy;

  if (memory_snap_to_grid == YES)
    memory_compute_tile_position(memory_x*memory_zoom, memory_y*memory_zoom);

  memory_copy_data_out(editor_data, memory_x, memory_y, editor_dx, editor_dy);
  common_copy_data_to_view(editor_data, editor_view, editor_dx, editor_dy, editor_zoom, editor_view_mode);
  memory_window_refresh();
  editor_window_set_title();
  tiled_window_refresh();

  gtk_drawing_area_size(GTK_DRAWING_AREA(editor_draw_area), editor_dx*editor_zoom, editor_dy*editor_zoom);
}


static void editor_view_n(GtkWidget *widget, gpointer data) {

  int i;


  i = (int)data;
  if (editor_view_mode == i)
    return;

  editor_view_mode = i;
  editor_window_refresh();
  editor_window_set_title();
}


static void editor_save(GtkWidget *widget, gpointer data) {

  if (editor_file_name_status == NO) {
    editor_save_as(widget, data);
    return;
  }

  png_save(editor_file_name, editor_dx, editor_dy, editor_data);
}


static void editor_save_as(GtkWidget *widget, gpointer data) {

  gtk_widget_show(editor_file_selection_save);
}


static void editor_open(GtkWidget *widget, gpointer data) {

  gtk_widget_show(editor_file_selection_open);
}


static void editor_file_save_ok(GtkWidget *widget, gpointer data) {

  char tmp[512];
  gchar *n;


  n = (gchar *)gtk_file_selection_get_filename(GTK_FILE_SELECTION(editor_file_selection_save));
  gtk_file_selection_set_filename(GTK_FILE_SELECTION(editor_file_selection_open), n);

  /* hide the file selection window */
  gtk_widget_hide(editor_file_selection_save);

  png_save(n, editor_dx, editor_dy, editor_data);

  strcpy(editor_file_name, n);
  editor_file_status = YES;
  editor_file_name_status = YES;

  /* strip the filename from the whole path */
  common_get_path(n, tmp);
  strcpy(editor_file_path, tmp);
}


static void editor_file_open_ok(GtkWidget *widget, gpointer data) {

  unsigned char *d;
  int dx, dy, bpp, i, l;
  gchar *n, tmp[512];


  /* backup the old data */
  editor_copy_to_undo();

  memory_title_reset(YES);

  n = (gchar *)gtk_file_selection_get_filename(GTK_FILE_SELECTION(editor_file_selection_open));

  /* hide the file selection window */
  gtk_widget_hide(editor_file_selection_open);

  /* find the file type */
  for (i = strlen(n)-1; i >= 0; i--) {
    if (n[i] == '.')
      break;
  }

  l = 0;
  if (string_compare_nocase(&n[i+1], "png") == 0)
    l |= 1;
  else if (string_compare_nocase(&n[i+1], "jpg") == 0 || string_compare_nocase(&n[i+1], "jpeg") == 0)
    l |= 2;
  else if (string_compare_nocase(&n[i+1], "bmp") == 0)
    l |= 4;
  else if (string_compare_nocase(&n[i+1], "pcx") == 0)
    l |= 8;
  else if (string_compare_nocase(&n[i+1], "tga") == 0)
    l |= 16;
  else
    l |= 31;

  i = FAILED;
  if (l & 1)
    i = png_load(n, &dx, &dy, &bpp, &d);
  if (l & 2 && i == FAILED)
    i = jpg_load(n, &dx, &dy, &bpp, &d);
  if (l & 4 && i == FAILED)
    i = bmp_load(n, &dx, &dy, &bpp, &d);
  if (l & 8 && i == FAILED)
    i = pcx_load(n, &dx, &dy, &bpp, &d);
  if (l & 16 && i == FAILED)
    i = tga_load(n, &dx, &dy, &bpp, &d);

  if (i == FAILED)
    return;

  /* reformat the image so it is of suitable format */
  common_reformat_image(&dx, &dy, &bpp, &d);

  strcpy(editor_file_name, n);
  editor_file_status = YES;

  /* strip the filename from the whole path */
  common_get_path(n, tmp);
  strcpy(editor_file_path, tmp);

  /* the filename can be used later only if the file was a png file */
  if (l == 1) {
    editor_file_name_status = YES;

    /* set the whole filename with path to the "save as" file selector */
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(editor_file_selection_save), n);
  }
  else {
    editor_file_name_status = NO;

    /* set the path to the "save as" file selector */
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(editor_file_selection_save), tmp);
  }

  /* copy it, or part of it, to the editor */
  common_copy_image(editor_data, editor_dx, editor_dy, d, dx, dy);
  free(d);

  /* copy to memory */
  memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

  /* refresh the draw areas */
  editor_window_refresh();
  memory_window_block_refresh();
  tiled_window_refresh();
}


static void editor_file_save_cancel(GtkWidget *widget, gpointer data) {

  gtk_widget_hide(editor_file_selection_save);
}


static void editor_file_open_cancel(GtkWidget *widget, gpointer data) {

  gtk_widget_hide(editor_file_selection_open);
}


static void editor_draw_fill(int x, int y) {

  unsigned char r, g, b, a;
  int i, n;


  memory_title_reset(YES);

  if (editor_mouse_button_mode == BG) {
    r = editor_pen_bg_r;
    g = editor_pen_bg_g;
    b = editor_pen_bg_b;
    a = editor_pen_bg_a;
  }
  else {
    r = editor_pen_r;
    g = editor_pen_g;
    b = editor_pen_b;
    a = editor_pen_a; 
  }

  i = 0;
  n = (y*editor_dx + x)<<2;

  if (editor_view_mode == VIEW_MODE_RGBA) {
    if (editor_data[n + 0] == editor_fill_r &&
	editor_data[n + 1] == editor_fill_g &&
	editor_data[n + 2] == editor_fill_b &&
	editor_data[n + 3] == editor_fill_a)
      i = 1;
  }
  else if (editor_view_mode == VIEW_MODE_RGB) {
    if (editor_data[n + 0] == editor_fill_r &&
	editor_data[n + 1] == editor_fill_g &&
	editor_data[n + 2] == editor_fill_b)
      i = 1;
  }
  else if (editor_view_mode == VIEW_MODE_ALPHA) {
    if (editor_data[n + 3] == editor_fill_a)
      i = 1;
  }

  if (i == 1) {
    /* fill the pixel */
    if (editor_view_mode == VIEW_MODE_RGBA) {
      editor_data[n + 0] = r;
      editor_data[n + 1] = g;
      editor_data[n + 2] = b;
      editor_data[n + 3] = a;
    }
    else if (editor_view_mode == VIEW_MODE_RGB) {
      editor_data[n + 0] = r;
      editor_data[n + 1] = g;
      editor_data[n + 2] = b;
    }
    else if (editor_view_mode == VIEW_MODE_ALPHA) {
      editor_data[n + 3] = a;
    }

    /* try to fill the neighboring pixels */
    if (x - 1 >= 0)
      editor_draw_fill(x - 1, y);
    if (y - 1 >= 0)
      editor_draw_fill(x, y - 1);
    if (x + 1 < editor_dx)
      editor_draw_fill(x + 1, y);
    if (y + 1 < editor_dy)
      editor_draw_fill(x, y + 1);
  }
}


static void editor_draw_pick(int x, int y) {

  unsigned char r, g, b, a;
  int i;


  i = (y*editor_dx + x)<<2;

  /* pick the color */
  r = editor_data[i + 0];
  g = editor_data[i + 1];
  b = editor_data[i + 2];
  a = editor_data[i + 3];

  /* change the pen to the picked color */
  if (editor_mouse_button_mode == BG) {
    if (editor_view_mode == VIEW_MODE_RGBA) {
      editor_pen_bg_r = r;
      editor_pen_bg_g = g;
      editor_pen_bg_b = b;
      editor_pen_bg_a = a;
    }
    else if (editor_view_mode == VIEW_MODE_RGB) {
      editor_pen_bg_r = r;
      editor_pen_bg_g = g;
      editor_pen_bg_b = b;
    }
    else {
      editor_pen_bg_a = a;
    }
  }
  else {
    if (editor_view_mode == VIEW_MODE_RGBA) {
      editor_pen_r = r;
      editor_pen_g = g;
      editor_pen_b = b;
      editor_pen_a = a;
    }
    else if (editor_view_mode == VIEW_MODE_RGB) {
      editor_pen_r = r;
      editor_pen_g = g;
      editor_pen_b = b;
    }
    else {
      editor_pen_a = a;
    }
  }

  palette_window_refresh();
}


static void editor_draw_pen(int x, int y) {

  unsigned char r, g, b, a;
  int i;


  if (editor_mouse_button_mode == BG) {
    r = editor_pen_bg_r;
    g = editor_pen_bg_g;
    b = editor_pen_bg_b;
    a = editor_pen_bg_a;
  }
  else {
    r = editor_pen_r;
    g = editor_pen_g;
    b = editor_pen_b;
    a = editor_pen_a; 
  }
  
  memory_title_reset(YES);

  i = (y*editor_dx + x)<<2;
  if (editor_view_mode == VIEW_MODE_RGBA) {
    editor_data[i + 0] = r;
    editor_data[i + 1] = g;
    editor_data[i + 2] = b;
    editor_data[i + 3] = a;
  }
  else if (editor_view_mode == VIEW_MODE_RGB) {
    editor_data[i + 0] = r;
    editor_data[i + 1] = g;
    editor_data[i + 2] = b;
  }
  else if (editor_view_mode == VIEW_MODE_ALPHA) {
    editor_data[i + 3] = a;
  }
}


static void editor_draw_blur_normal(unsigned int x, unsigned int y) {

  unsigned i, a, b;
  float cr, cg, cb, ca;


  memory_title_reset(YES);

  cr = cg = cb = ca = 0.0f;

  if (editor_view_mode == VIEW_MODE_RGBA) {
    for (a = 0; a < 3; a++) {
      for (b = 0; b < 3; b++) {
	i = (((y+a-1)%editor_dy)*editor_dx + ((x+b-1)%editor_dx))<<2;
	cr += editor_data[i + 0];
	cg += editor_data[i + 1];
	cb += editor_data[i + 2];
	ca += editor_data[i + 3];
      }
    }
    editor_data[(y*editor_dx + x)*4 + 0] = cr/9.0f;
    editor_data[(y*editor_dx + x)*4 + 1] = cg/9.0f;
    editor_data[(y*editor_dx + x)*4 + 2] = cb/9.0f;
    editor_data[(y*editor_dx + x)*4 + 3] = ca/9.0f;
  }
  else if (editor_view_mode == VIEW_MODE_RGB) {
    for (a = 0; a < 3; a++) {
      for (b = 0; b < 3; b++) {
	i = (((y+a-1)%editor_dy)*editor_dx + ((x+b-1)%editor_dx))<<2;
	cr += editor_data[i + 0];
	cg += editor_data[i + 1];
	cb += editor_data[i + 2];
      }
    }
    editor_data[(y*editor_dx + x)*4 + 0] = cr/9.0f;
    editor_data[(y*editor_dx + x)*4 + 1] = cg/9.0f;
    editor_data[(y*editor_dx + x)*4 + 2] = cb/9.0f;
  }
  else if (editor_view_mode == VIEW_MODE_ALPHA) {
    for (a = 0; a < 3; a++) {
      for (b = 0; b < 3; b++) {
	i = (((y+a-1)%editor_dy)*editor_dx + ((x+b-1)%editor_dx))<<2;
	ca += editor_data[i + 3];
      }
    }
    editor_data[(y*editor_dx + x)*4 + 3] = ca/9.0f;
  }
}


static void editor_draw_blur_gaussian(unsigned int x, unsigned int y) {

  unsigned i, a, b, n;
  float cr, cg, cb, ca;
  float filter[3*3] = { 1.0/36.0, 1.0/9.0, 1.0/36.0,
			1.0/9.0, 4.0/9.0, 1.0/9.0,
			1.0/36.0, 1.0/9.0, 1.0/36.0 };


  memory_title_reset(YES);

  cr = cg = cb = ca = 0.0f;

  if (editor_view_mode == VIEW_MODE_RGBA) {
    for (n = 0, a = 0; a < 3; a++) {
      for (b = 0; b < 3; b++, n++) {
	i = (((y+a-1)%editor_dy)*editor_dx + ((x+b-1)%editor_dx))<<2;
	cr += editor_data[i + 0]*filter[n];
	cg += editor_data[i + 1]*filter[n];
	cb += editor_data[i + 2]*filter[n];
	ca += editor_data[i + 3]*filter[n];
      }
    }
    editor_data[(y*editor_dx + x)*4 + 0] = cr;
    editor_data[(y*editor_dx + x)*4 + 1] = cg;
    editor_data[(y*editor_dx + x)*4 + 2] = cb;
    editor_data[(y*editor_dx + x)*4 + 3] = ca;
  }
  else if (editor_view_mode == VIEW_MODE_RGB) {
    for (n = 0, a = 0; a < 3; a++) {
      for (b = 0; b < 3; b++, n++) {
	i = (((y+a-1)%editor_dy)*editor_dx + ((x+b-1)%editor_dx))<<2;
	cr += editor_data[i + 0]*filter[n];
	cg += editor_data[i + 1]*filter[n];
	cb += editor_data[i + 2]*filter[n];
      }
    }
    editor_data[(y*editor_dx + x)*4 + 0] = cr;
    editor_data[(y*editor_dx + x)*4 + 1] = cg;
    editor_data[(y*editor_dx + x)*4 + 2] = cb;
  }
  else if (editor_view_mode == VIEW_MODE_ALPHA) {
    for (n = 0, a = 0; a < 3; a++) {
      for (b = 0; b < 3; b++, n++) {
	i = (((y+a-1)%editor_dy)*editor_dx + ((x+b-1)%editor_dx))<<2;
	ca += editor_data[i + 3]*filter[n];
      }
    }
    editor_data[(y*editor_dx + x)*4 + 3] = ca;
  }
}


static void editor_draw_cursor(int x, int y, GtkWidget *widget) {

  unsigned char r, g, b, a;
  int i;


  /* backup the old pixel */
  i = (y*editor_dx + x)<<2;
  r = editor_data[i + 0];
  g = editor_data[i + 1];
  b = editor_data[i + 2];
  a = editor_data[i + 3];

  /* highlight the pixel */
  editor_data[i + 0] = editor_pen_r;
  editor_data[i + 1] = editor_pen_g;
  editor_data[i + 2] = editor_pen_b;
  editor_data[i + 3] = editor_pen_a;

  /* display it */
  common_copy_pixel_to_view(editor_data, editor_view, editor_dx, editor_dy, x, y, editor_zoom, editor_view_mode);

  /* return the old value */
  editor_data[i + 0] = r;
  editor_data[i + 1] = g;
  editor_data[i + 2] = b;
  editor_data[i + 3] = a;

  /* refresh the editor draw area */
  gdk_draw_rgb_image(widget->window, widget->style->fg_gc[GTK_STATE_NORMAL], x*editor_zoom, y*editor_zoom,
		     editor_zoom, editor_zoom, GDK_RGB_DITHER_MAX, editor_view + y*editor_zoom*editor_dx*3*editor_zoom + x*3*editor_zoom, editor_dx*editor_zoom*3);
}


static int editor_need_to_draw(int x, int y) {

  unsigned char r, g, b, a;
  int i;


  if (editor_mouse_button_mode == BG) {
    r = editor_pen_bg_r;
    g = editor_pen_bg_g;
    b = editor_pen_bg_b;
    a = editor_pen_bg_a;
  }
  else {
    r = editor_pen_r;
    g = editor_pen_g;
    b = editor_pen_b;
    a = editor_pen_a; 
  }

  i = (y*editor_dx + x)<<2;
  if (editor_view_mode == VIEW_MODE_RGBA) {
    if (r == editor_data[i + 0] &&
	g == editor_data[i + 1] &&
	b == editor_data[i + 2] &&
	a == editor_data[i + 3])
      return NO;
    return YES;
  }
  else if (editor_view_mode == VIEW_MODE_RGB) {
    if (r == editor_data[i + 0] &&
	g == editor_data[i + 1] &&
	b == editor_data[i + 2])
      return NO;
    return YES;
  }
  else if (editor_view_mode == VIEW_MODE_ALPHA) {
    if (a == editor_data[i + 3])
      return NO;
    return YES;
  }

  return YES;
}


/* old mouse x and y */
static int mouse_old_x = -1, mouse_old_y = -1;


static gint editor_button_press(GtkWidget *widget, GdkEventButton *event) {

  int x, y, i;
  

  editor_mouse_button = ON;

  if (event->button == 3)
    editor_mouse_button_mode = BG;
  else
    editor_mouse_button_mode = FG;
  
  x = event->x/editor_zoom;
  y = event->y/editor_zoom;

  if (y < 0 || y >= editor_dy || x < 0 || x >= editor_dy)
    return FALSE;

  mouse_old_x = x;
  mouse_old_y = y;

  if (editor_tool_mode == TOOL_MODE_FILL) {
    if (editor_need_to_draw(x, y) == NO) {
      /* refresh the pointer */
      editor_draw_cursor(x, y, widget);

      return FALSE;
    }

    i = (y*editor_dx + x)<<2;
    editor_fill_r = editor_data[i + 0];
    editor_fill_g = editor_data[i + 1];
    editor_fill_b = editor_data[i + 2];
    editor_fill_a = editor_data[i + 3];

    /* backup the old data */
    editor_copy_to_undo();

    editor_draw_fill(x, y);

    /* copy to memory */
    memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

    /* refresh the draw areas */
    editor_window_refresh();
    memory_window_block_refresh();
    tiled_window_refresh();

    /* refresh the pointer */
    editor_draw_cursor(x, y, widget);
  }
  else if (editor_tool_mode == TOOL_MODE_BLUR_GAUSSIAN || editor_tool_mode == TOOL_MODE_BLUR_NORMAL) {
    /* backup the old data */
    editor_copy_to_undo();

    if (editor_tool_mode == TOOL_MODE_BLUR_GAUSSIAN)
      editor_draw_blur_gaussian((unsigned int)x, (unsigned int)y);
    else
      editor_draw_blur_normal((unsigned int)x, (unsigned int)y);

    /* refresh the editor draw area */
    common_copy_pixel_to_view(editor_data, editor_view, editor_dx, editor_dy, x, y, editor_zoom, editor_view_mode);

    /* update the other windows, too, if required */
    if (tiled_real_time == YES)
      tiled_window_refresh();
    if (memory_real_time == YES) {
      /* copy to memory */
      memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);
      memory_window_block_refresh();
    }

    gtk_widget_queue_draw(editor_draw_area);
  }
  else if (editor_tool_mode == TOOL_MODE_PEN) {
    /* backup the old data */
    editor_copy_to_undo();

    if (editor_need_to_draw(x, y) == NO) {
      /* update the previous pixel */
      common_copy_pixel_to_view(editor_data, editor_view, editor_dx, editor_dy, mouse_old_x, mouse_old_y, editor_zoom, editor_view_mode);
      gdk_draw_rgb_image(widget->window, widget->style->fg_gc[GTK_STATE_NORMAL], mouse_old_x*editor_zoom, mouse_old_y*editor_zoom,
			 editor_zoom, editor_zoom, GDK_RGB_DITHER_MAX, editor_view + mouse_old_y*editor_zoom*editor_dx*3*editor_zoom + mouse_old_x*3*editor_zoom, editor_dx*editor_zoom*3);

      return FALSE;
    }

    editor_draw_pen(x, y);

    /* refresh the editor draw area */
    common_copy_pixel_to_view(editor_data, editor_view, editor_dx, editor_dy, x, y, editor_zoom, editor_view_mode);

    /* update the other windows, too, if required */
    if (tiled_real_time == YES)
      tiled_window_refresh();
    if (memory_real_time == YES) {
      /* copy to memory */
      memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);
      memory_window_block_refresh();
    }

    gtk_widget_queue_draw(editor_draw_area);
  }
  else if (editor_tool_mode == TOOL_MODE_PICK) {
    editor_draw_pick(x, y);

    /* refresh the pointer */
    editor_draw_cursor(x, y, widget);
  }

  return FALSE;
}


static gint editor_button_release(GtkWidget *widget, GdkEventButton *event) {

  editor_mouse_button = OFF;

  if (mouse_old_x >= 0 && mouse_old_y >= 0)
    editor_draw_cursor(mouse_old_x, mouse_old_y, widget);

  /* copy to memory */
  memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);

  /* update the other windows only after the mouse button has been released */
  memory_window_block_refresh();
  tiled_window_refresh();

  return FALSE;
}


static gint editor_motion_notify(GtkWidget *widget, GdkEventButton *event) {

  int x, y;


  x = event->x/editor_zoom;
  y = event->y/editor_zoom;

  if (y < 0 || y >= editor_dy || x < 0 || x >= editor_dx)
    return FALSE;

  /* do we need to update the pixel? */
  if (mouse_old_x == x && mouse_old_y == y)
    return FALSE;

  if (editor_mouse_button == OFF) {
    /* no button pressed -> highlight the pixel under the pointer */

    /* update the previous pixel */
    if (mouse_old_x >= 0 && mouse_old_y >= 0) {
      common_copy_pixel_to_view(editor_data, editor_view, editor_dx, editor_dy, mouse_old_x, mouse_old_y, editor_zoom, editor_view_mode);
      gdk_draw_rgb_image(widget->window, widget->style->fg_gc[GTK_STATE_NORMAL], mouse_old_x*editor_zoom, mouse_old_y*editor_zoom,
			 editor_zoom, editor_zoom, GDK_RGB_DITHER_MAX, editor_view + mouse_old_y*editor_zoom*editor_dx*3*editor_zoom + mouse_old_x*3*editor_zoom, editor_dx*editor_zoom*3);
    }

    editor_draw_cursor(x, y, widget);

    mouse_old_x = x;
    mouse_old_y = y;

    return FALSE;
  }

  if (editor_tool_mode == TOOL_MODE_PEN) {
    /* do we need to update the pixel? */
    if (editor_need_to_draw(x, y) == YES) {
      editor_draw_pen(x, y);

      /* refresh the editor draw area */
      common_copy_pixel_to_view(editor_data, editor_view, editor_dx, editor_dy, x, y, editor_zoom, editor_view_mode);
      gdk_draw_rgb_image(widget->window, widget->style->fg_gc[GTK_STATE_NORMAL], x*editor_zoom, y*editor_zoom,
			 editor_zoom, editor_zoom, GDK_RGB_DITHER_MAX, editor_view + y*editor_zoom*editor_dx*3*editor_zoom + x*3*editor_zoom, editor_dx*editor_zoom*3);

      /* update the other windows, too, if required */
      if (tiled_real_time == YES)
	tiled_window_refresh();
      if (memory_real_time == YES) {
	/* copy to memory */
	memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);
	memory_window_block_refresh();
      }
    }
  }
  else if (editor_tool_mode == TOOL_MODE_BLUR_GAUSSIAN || editor_tool_mode == TOOL_MODE_BLUR_NORMAL) {

    if (editor_tool_mode == TOOL_MODE_BLUR_GAUSSIAN)
      editor_draw_blur_gaussian((unsigned int)x, (unsigned int)y);
    else
      editor_draw_blur_normal((unsigned int)x, (unsigned int)y);

    /* refresh the editor draw area */
    common_copy_pixel_to_view(editor_data, editor_view, editor_dx, editor_dy, x, y, editor_zoom, editor_view_mode);
    gdk_draw_rgb_image(widget->window, widget->style->fg_gc[GTK_STATE_NORMAL], x*editor_zoom, y*editor_zoom,
		       editor_zoom, editor_zoom, GDK_RGB_DITHER_MAX, editor_view + y*editor_zoom*editor_dx*3*editor_zoom + x*3*editor_zoom, editor_dx*editor_zoom*3);

    /* update the other windows, too, if required */
    if (tiled_real_time == YES)
      tiled_window_refresh();
    if (memory_real_time == YES) {
      /* copy to memory */
      memory_copy_data_in(editor_data, memory_x, memory_y, editor_dx, editor_dy);
      memory_window_block_refresh();
    }
  }
  else if (editor_tool_mode == TOOL_MODE_PICK) {
    editor_draw_pick(x, y);

    /* update the previous pixel */
    if (mouse_old_x >= 0 && mouse_old_y >= 0) {
      common_copy_pixel_to_view(editor_data, editor_view, editor_dx, editor_dy, mouse_old_x, mouse_old_y, editor_zoom, editor_view_mode);
      gdk_draw_rgb_image(widget->window, widget->style->fg_gc[GTK_STATE_NORMAL], mouse_old_x*editor_zoom, mouse_old_y*editor_zoom,
			 editor_zoom, editor_zoom, GDK_RGB_DITHER_MAX, editor_view + mouse_old_y*editor_zoom*editor_dx*3*editor_zoom + mouse_old_x*3*editor_zoom, editor_dx*editor_zoom*3);
    }

    /* refresh the pointer */
    editor_draw_cursor(x, y, widget);
  }
  else if (editor_tool_mode == TOOL_MODE_FILL) {
    /* update the previous pixel */
    if (mouse_old_x >= 0 && mouse_old_y >= 0) {
      common_copy_pixel_to_view(editor_data, editor_view, editor_dx, editor_dy, mouse_old_x, mouse_old_y, editor_zoom, editor_view_mode);
      gdk_draw_rgb_image(widget->window, widget->style->fg_gc[GTK_STATE_NORMAL], mouse_old_x*editor_zoom, mouse_old_y*editor_zoom,
			 editor_zoom, editor_zoom, GDK_RGB_DITHER_MAX, editor_view + mouse_old_y*editor_zoom*editor_dx*3*editor_zoom + mouse_old_x*3*editor_zoom, editor_dx*editor_zoom*3);
    }

    /* refresh the pointer */
    editor_draw_cursor(x, y, widget);
  }

  mouse_old_x = x;
  mouse_old_y = y;

  return FALSE;
}


static gint editor_leave_notify(GtkWidget *widget, GdkEventButton *event) {

  if (editor_view == NULL)
    return FALSE;

  /* update the previous pixel */
  if (mouse_old_x >= 0 && mouse_old_y >= 0) {
    common_copy_pixel_to_view(editor_data, editor_view, editor_dx, editor_dy, mouse_old_x, mouse_old_y, editor_zoom, editor_view_mode);
    gdk_draw_rgb_image(widget->window, widget->style->fg_gc[GTK_STATE_NORMAL], mouse_old_x*editor_zoom, mouse_old_y*editor_zoom,
		       editor_zoom, editor_zoom, GDK_RGB_DITHER_MAX, editor_view + mouse_old_y*editor_zoom*editor_dx*3*editor_zoom + mouse_old_x*3*editor_zoom, editor_dx*editor_zoom*3);
  }

  mouse_old_x = -1;
  mouse_old_y = -1;

  return FALSE;
}


static void editor_tools_n(GtkWidget *widget, gpointer data) {

  editor_tool_mode = (int)data;

  editor_window_set_title();
}


static gboolean editor_draw_area_expose(GtkWidget *widget, GdkEventExpose *event, gpointer user_data) {

  if (editor_view != NULL) {
    gdk_draw_rgb_image(widget->window, widget->style->fg_gc[GTK_STATE_NORMAL], 0, 0,
		       editor_dx*editor_zoom, editor_dy*editor_zoom, GDK_RGB_DITHER_MAX, editor_view, editor_dx*editor_zoom*3);
  }

  return FALSE;
}
