
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gtk/gtkadjustment.h>
#include <gtk/gtkwidget.h>
#include <gdk/gdkkeysyms.h>

#include "defines.h"
#include "memory.h"
#include "editor.h"
#include "common.h"
#include "tiled.h"
#include "png.h"


/* internal functions */
static void tiled_window_set_title(void);
static void tiled_zoom_n(GtkWidget *widget, gpointer data);
static void tiled_view_n(GtkWidget *widget, gpointer data);
static void tiled_size_n(GtkWidget *widget, gpointer data);
static gint tiled_hide_window(GtkWidget *widget, gpointer data);
static void tiled_update_mode(GtkWidget *widget, gpointer data);
static gboolean tiled_draw_area_expose(GtkWidget *widget, GdkEventExpose *event, gpointer user_data);


/* tiled window data (GTK) */
GtkWidget *tiled_window = NULL;
GtkWidget *tiled_vbox1 = NULL, *tiled_vbox2 = NULL, *tiled_hbox1 = NULL;
GtkWidget *tiled_menubar = NULL;
GtkWidget *tiled_draw_area = NULL;
GtkWidget *tiled_file_selection_save = NULL, *tiled_file_selection_open = NULL;

GtkAccelGroup *tiled_accel_group = NULL;
GtkItemFactory *tiled_item_factory = NULL;

GtkItemFactoryEntry tiled_menu_items[] = {
  { "/_Zoom",       NULL,         NULL, 0, "<Branch>" },
  { "/Zoom/100%",   NULL,         tiled_zoom_n, 1, "<RadioItem>" },
  { "/Zoom/200%",   NULL,         tiled_zoom_n, 2, "/Zoom/100%" },
  { "/Zoom/300%",   NULL,         tiled_zoom_n, 3, "/Zoom/100%" },
  { "/Zoom/400%",   NULL,         tiled_zoom_n, 4, "/Zoom/100%" },
  { "/Zoom/500%",   NULL,         tiled_zoom_n, 5, "/Zoom/100%" },
  { "/Zoom/1000%",  NULL,         tiled_zoom_n, 10, "/Zoom/100%" },
  { "/_Size",       NULL,         NULL, 0, "<Branch>" },
  { "/Size/1x1",    NULL,         tiled_size_n, 1, "<RadioItem>" },
  { "/Size/3x3",    NULL,         tiled_size_n, 3, "/Size/1x1" },
  { "/Size/5x5",    NULL,         tiled_size_n, 5, "/Size/1x1" },
  { "/Size/7x7",    NULL,         tiled_size_n, 7, "/Size/1x1" },
  { "/Size/9x9",    NULL,         tiled_size_n, 9, "/Size/1x1" },
  { "/_View",       NULL,         NULL, 0, "<Branch>" },
  { "/View/RGBA",   "<alt>6",     tiled_view_n, VIEW_MODE_RGBA, "<RadioItem>" },
  { "/View/RGB",    "<alt>7",     tiled_view_n, VIEW_MODE_RGB, "/View/RGBA" },
  { "/View/Alpha",  "<alt>8",     tiled_view_n, VIEW_MODE_ALPHA, "/View/RGBA" },
  { "/View/sep1",   NULL,         NULL, 0, "<Separator>" },
  { "/View/Updated in Real-Time", NULL, tiled_update_mode, 0, "<ToggleItem>" },
};

/* tiled data */
int tiled_zoom = 2;
int tiled_dx = 0, tiled_dy = 0;
int tiled_tiles_dx = 3, tiled_tiles_dy = 3;
unsigned char *tiled_data = NULL, *tiled_view = NULL;

/* tiled view status */
int tiled_view_mode = VIEW_MODE_RGBA;
int tiled_view_status = ON;

/* real time updates */
int tiled_real_time = NO;



int tiled_window_init(void) {

  /* copy data from editor window */
  tiled_dx = editor_dx*tiled_tiles_dx;
  tiled_dy = editor_dy*tiled_tiles_dy;

  /* window */
  tiled_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  tiled_window_set_title();

  /* menu */
  tiled_accel_group = gtk_accel_group_new();
  tiled_item_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", tiled_accel_group);
  gtk_item_factory_create_items(tiled_item_factory, sizeof(tiled_menu_items)/sizeof(tiled_menu_items[0]), tiled_menu_items, NULL);
  gtk_window_add_accel_group(GTK_WINDOW(tiled_window), tiled_accel_group);
  tiled_menubar = gtk_item_factory_get_widget(tiled_item_factory, "<main>");

  /* drawing area */
  tiled_draw_area = gtk_drawing_area_new();
  gtk_drawing_area_size(GTK_DRAWING_AREA(tiled_draw_area), tiled_dx*tiled_zoom, tiled_dy*tiled_zoom);

  /* boxes */
  tiled_vbox1 = gtk_vbox_new(FALSE, 0);
  tiled_vbox2 = gtk_vbox_new(FALSE, 0);
  tiled_hbox1 = gtk_hbox_new(FALSE, 0);
  gtk_container_border_width(GTK_CONTAINER(tiled_vbox1), 1);
  gtk_container_border_width(GTK_CONTAINER(tiled_vbox2), 1);
  gtk_container_border_width(GTK_CONTAINER(tiled_hbox1), 1);

  /* place boxes and widgets */
  gtk_container_add(GTK_CONTAINER(tiled_window), tiled_vbox1);
  gtk_box_pack_start(GTK_BOX(tiled_vbox1), tiled_menubar, FALSE, FALSE, 0);
  gtk_container_add(GTK_CONTAINER(tiled_vbox1), tiled_hbox1);
  gtk_box_pack_start(GTK_BOX(tiled_hbox1), tiled_vbox2, TRUE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(tiled_vbox2), tiled_draw_area, TRUE, FALSE, 0);

  /* signals */
  gtk_signal_connect(GTK_OBJECT(tiled_draw_area), "expose_event", GTK_SIGNAL_FUNC(tiled_draw_area_expose), NULL);
  gtk_signal_connect(GTK_OBJECT(tiled_window), "delete_event", GTK_SIGNAL_FUNC(tiled_hide_window), NULL);
  gtk_signal_connect(GTK_OBJECT(tiled_window), "destroy", GTK_SIGNAL_FUNC(tiled_hide_window), NULL);

  gtk_window_set_resizable(GTK_WINDOW(tiled_window), FALSE);

  return SUCCEEDED;
}


void tiled_resize_view(int zoom) {

  tiled_view = realloc(tiled_view, tiled_dx*zoom*tiled_dy*zoom*3);
  if (tiled_view == NULL) {
    fprintf(stderr, "TILED_RESIZE_VIEW: Out of tiled error.\n");
    return;
  }

  tiled_zoom = zoom;
}


void tiled_resize_data(int dx, int dy) {

  tiled_data = realloc(tiled_data, dx*dy*4);
  if (tiled_data == NULL) {
    fprintf(stderr, "TILED_RESIZE_DATA: Out of tiled error.\n");
    return;
  }

  memset(tiled_data, 255, dx*dy*4);

  tiled_dx = dx;
  tiled_dy = dy;
}


static void tiled_window_set_title(void) {

  char tmp[512];


  if (tiled_window == NULL)
    return;

  sprintf(tmp, "Tiled View | %dx%d | ", tiled_tiles_dx, tiled_tiles_dy);

  if (tiled_view_mode == VIEW_MODE_RGBA)
    strcat(tmp, "RGBA");
  else if (tiled_view_mode == VIEW_MODE_RGB)
    strcat(tmp, "RGB");
  else if (tiled_view_mode == VIEW_MODE_ALPHA)
    strcat(tmp, "Alpha");

  gtk_window_set_title(GTK_WINDOW(tiled_window), tmp);
}


static void tiled_zoom_n(GtkWidget *widget, gpointer data) {

  int zoom;


  zoom = (int)data;
  if (tiled_zoom == zoom)
    return;

  tiled_resize_view(zoom);
  common_copy_data_to_view(tiled_data, tiled_view, tiled_dx, tiled_dy, tiled_zoom, tiled_view_mode);

  gtk_drawing_area_size(GTK_DRAWING_AREA(tiled_draw_area), tiled_dx*tiled_zoom, tiled_dy*tiled_zoom);
}


static void tiled_view_n(GtkWidget *widget, gpointer data) {

  int i;


  i = (int)data;
  if (tiled_view_mode == i)
    return;

  tiled_view_mode = i;
  tiled_window_refresh();
  tiled_window_set_title();
}


int tiled_window_import_data(void) {

  int x, y, a, b, i, n, m, o, p;


  for (y = 0; y < tiled_tiles_dy; y++) {
    n = (y*editor_dy*tiled_dx)<<2;
    for (x = 0; x < tiled_tiles_dx; x++) {
      m = ((x*editor_dx)<<2) + n;
      for (i = 0, a = 0; a < editor_dy; a++) {
	o = ((a*tiled_dx)<<2) + m;
	p = o;
	for (b = 0; b < editor_dx; b++) {
	  tiled_data[p++] = editor_data[i++];
	  tiled_data[p++] = editor_data[i++];
	  tiled_data[p++] = editor_data[i++];
	  tiled_data[p++] = editor_data[i++];
	}
      }
    }
  }

  return SUCCEEDED;
}


int tiled_window_refresh(void) {

  if (tiled_data == NULL || tiled_view == NULL)
    return FAILED;

  if (tiled_view_status == OFF)
    return FAILED;

  /* tiled window */
  if (tiled_dx != editor_dx*tiled_tiles_dx || tiled_dy != editor_dy*tiled_tiles_dy) {
    tiled_resize_data(editor_dx*tiled_tiles_dx, editor_dy*tiled_tiles_dy);
    tiled_resize_view(tiled_zoom);
    tiled_window_import_data();
    common_copy_data_to_view(tiled_data, tiled_view, tiled_dx, tiled_dy, tiled_zoom, tiled_view_mode);
    gtk_drawing_area_size(GTK_DRAWING_AREA(tiled_draw_area), tiled_dx*tiled_zoom, tiled_dy*tiled_zoom);
    return SUCCEEDED;
  }

  tiled_window_import_data();
  common_copy_data_to_view(tiled_data, tiled_view, tiled_dx, tiled_dy, tiled_zoom, tiled_view_mode);

  /* refresh the tiled draw area */
  gtk_widget_queue_draw(tiled_draw_area);

  return SUCCEEDED;
}


static void tiled_size_n(GtkWidget *widget, gpointer data) {

  int i;


  i = (int)data;
  if (tiled_tiles_dx == i && tiled_tiles_dy == i)
    return;

  tiled_tiles_dx = i;
  tiled_tiles_dy = i;

  tiled_window_refresh();
  tiled_window_set_title();
}


static gint tiled_hide_window(GtkWidget *widget, gpointer data) {

  gtk_check_menu_item_set_active((GtkCheckMenuItem *)gtk_item_factory_get_widget(memory_item_factory, "/Windows/Tiled View"), FALSE);

  return TRUE;
}


static void tiled_update_mode(GtkWidget *widget, gpointer data) {

  if (tiled_real_time == NO)
    tiled_real_time = YES;
  else
    tiled_real_time = NO;
}


static gboolean tiled_draw_area_expose(GtkWidget *widget, GdkEventExpose *event, gpointer user_data) {

  if (tiled_view != NULL) {
    gdk_draw_rgb_image(widget->window, widget->style->fg_gc[GTK_STATE_NORMAL], 0, 0,
		       tiled_dx*tiled_zoom, tiled_dy*tiled_zoom, GDK_RGB_DITHER_MAX, tiled_view, tiled_dx*tiled_zoom*3);
  }

  return FALSE;
}
